# Copyright 2025 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Common tests for AbstractCheckpointManager subclasses."""

# pylint: disable=protected-access
import copy
from typing import Any, Callable, Sequence
import unittest
from unittest import mock

from absl.testing import parameterized
from etils import epath
from flax import linen as nn
from flax.training import train_state
import jax
from jax.experimental import pjit
import jax.numpy as jnp
import numpy as np
import optax
from orbax.checkpoint import args
from orbax.checkpoint import test_utils
from orbax.checkpoint import utils
from orbax.checkpoint._src import asyncio_utils
from orbax.checkpoint._src.checkpointers import async_checkpointer
from orbax.checkpoint._src.checkpointers import checkpointer as checkpointer_lib
from orbax.checkpoint._src.handlers import async_checkpoint_handler
from orbax.checkpoint._src.handlers import composite_checkpoint_handler
from orbax.checkpoint._src.handlers import pytree_checkpoint_handler
from orbax.checkpoint._src.logging import step_statistics
from orbax.checkpoint._src.metadata import array_metadata
from orbax.checkpoint._src.metadata import array_metadata_store as array_metadata_store_lib
from orbax.checkpoint._src.metadata import checkpoint
from orbax.checkpoint._src.metadata import step_metadata_serialization
from orbax.checkpoint._src.metadata import tree as tree_metadata
from orbax.checkpoint._src.multihost import multihost
from orbax.checkpoint._src.path import atomicity
from orbax.checkpoint._src.path import step
from orbax.checkpoint._src.serialization import serialization
from orbax.checkpoint._src.serialization import type_handlers
from orbax.checkpoint._src.testing import test_tree_utils

from .pyglib import gfile


PyTreeCheckpointHandler = pytree_checkpoint_handler.PyTreeCheckpointHandler
AsyncCheckpointHandler = async_checkpoint_handler.AsyncCheckpointHandler
AsyncCheckpointer = async_checkpointer.AsyncCheckpointer
ArrayRestoreArgs = pytree_checkpoint_handler.ArrayRestoreArgs
TrainState = train_state.TrainState
Mesh = jax.sharding.Mesh
PyTree = Any
PLACEHOLDER = type_handlers.PLACEHOLDER


class CheckpointerTestBase:
  """Common tests for AbstractCheckpointer subclasses."""

  class Test(parameterized.TestCase, unittest.IsolatedAsyncioTestCase):
    """Structure allows test to run as subclasses, not base class."""

    def checkpointer(self, handler, **kwargs) -> checkpointer_lib.Checkpointer:
      raise NotImplementedError

    def setUp(self):
      super().setUp()
      pytree, mesh_tree, axes_tree = test_utils.setup_sharded_pytree()
      doubled_pytree = test_utils.apply_function(pytree, lambda x: x * 2)

      self.empty_pytree = jax.tree.map(
          lambda x: object(), pytree, is_leaf=test_utils.is_leaf
      )
      self.pytree = pytree
      self.doubled_pytree = doubled_pytree
      self.mesh_tree = mesh_tree
      self.axes_tree = axes_tree
      self.pytree_restore_args = jax.tree.map(
          lambda mesh, axes: ArrayRestoreArgs(mesh=mesh, mesh_axes=axes),
          self.mesh_tree,
          self.axes_tree,
      )
      self.directory = (
          epath.Path(self.create_tempdir(name='checkpointing_test').full_path)
          / 'ckpt'
      )
      test_utils.set_tensorstore_driver_for_test()

      test_utils.sync_global_processes('CheckpointerTest:setup_complete')

    def tearDown(self):
      test_utils.sync_global_processes('CheckpointerTest:tests_complete')
      super().tearDown()

    def wait_if_async(self, checkpointer):
      if isinstance(checkpointer, AsyncCheckpointer):
        checkpointer.wait_until_finished()

    def is_world_readable(self, path):
      if not gfile.Exists(path):
        return False
      mode = gfile.Stat(path, stat_proto=True).mode
      return (mode & 0o770) != mode

    def test_save_restore(self):
      """Basic save and restore test."""
      with self.checkpointer(PyTreeCheckpointHandler()) as checkpointer:
        checkpointer.save(self.directory, self.pytree)
        self.wait_if_async(checkpointer)
        restored = checkpointer.restore(
            self.directory, restore_args=self.pytree_restore_args
        )
        test_utils.assert_tree_equal(self, self.pytree, restored)

    def test_save_restore_with_args(self):
      """Basic save and restore test."""
      with self.checkpointer(PyTreeCheckpointHandler()) as checkpointer:
        checkpointer.save(self.directory, args=args.PyTreeSave(self.pytree))
        self.wait_if_async(checkpointer)
        restored = checkpointer.restore(
            self.directory,
            args=args.PyTreeRestore(restore_args=self.pytree_restore_args),
        )
        test_utils.assert_tree_equal(self, self.pytree, restored)

    def test_save_restore_no_kwargs(self):
      """Restore with no GDA args."""
      with self.checkpointer(PyTreeCheckpointHandler()) as checkpointer:
        expected = jax.tree.map(test_utils.replicate_sharded_array, self.pytree)
        expected = jax.tree.map(
            lambda x: np.asarray(x.addressable_data(0)), expected
        )
        checkpointer.save(self.directory, expected)
        self.wait_if_async(checkpointer)
        restored = checkpointer.restore(self.directory)
        test_utils.assert_tree_equal(self, expected, restored)

    def test_restore_missing_path(self):
      """Restore with invalid item."""
      with self.checkpointer(PyTreeCheckpointHandler()) as checkpointer:
        with self.assertRaises(FileNotFoundError):
          checkpointer.restore('path/to/missing')

    def test_no_overwrite_existing(self):
      """Test same step does not overwrite."""
      with self.checkpointer(PyTreeCheckpointHandler()) as checkpointer:
        checkpointer.save(self.directory, self.pytree)
        self.wait_if_async(checkpointer)
        with self.assertRaises(ValueError):
          checkpointer.save(self.directory, self.doubled_pytree)
        self.wait_if_async(checkpointer)
        restored = checkpointer.restore(
            self.directory, restore_args=self.pytree_restore_args
        )
        expected = self.pytree
        test_utils.assert_tree_equal(self, expected, restored)
        self.assertFalse(self.is_world_readable(self.directory))

    def test_overwrite_existing(self):
      """Test overwrite existing path."""
      with self.checkpointer(PyTreeCheckpointHandler()) as checkpointer:
        checkpointer.save(self.directory, self.pytree)
        self.wait_if_async(checkpointer)
        checkpointer.save(self.directory, self.doubled_pytree, force=True)
        self.wait_if_async(checkpointer)
        restored = checkpointer.restore(
            self.directory, restore_args=self.pytree_restore_args
        )
        expected = self.doubled_pytree
        test_utils.assert_tree_equal(self, expected, restored)
        self.assertFalse(self.is_world_readable(self.directory))

    def test_flax_train_state(self):
      """Test using flax model."""

      class MLP(nn.Module):
        """A simple MLP model."""

        @nn.compact
        def __call__(self, x: jnp.ndarray) -> jnp.ndarray:
          x = x.reshape((x.shape[0], -1))  # flatten
          x = nn.Dense(features=8)(x)
          return x

      model = MLP()
      mesh = Mesh(np.asarray(jax.devices()), ('devices',))
      mesh_axes = jax.sharding.PartitionSpec()

      @jax.jit
      def init_state():
        params = model.init(jax.random.PRNGKey(0), jnp.ones([8, 8]))
        tx = optax.adamw(learning_rate=0.001)
        state = TrainState.create(apply_fn=model.apply, params=params, tx=tx)
        return state

      init = pjit.pjit(init_state, in_shardings=None, out_shardings=mesh_axes)

      with Mesh(mesh.devices, mesh.axis_names):
        state = init()
        state_shape = jax.eval_shape(init)

      restore_args = jax.tree.map(
          lambda _: ArrayRestoreArgs(mesh=mesh, mesh_axes=mesh_axes),
          state_shape,
      )

      with self.checkpointer(PyTreeCheckpointHandler()) as checkpointer:
        checkpointer.save(self.directory, state)
        self.wait_if_async(checkpointer)
        # Already fully replicated, don't need to provide args.
        restored = checkpointer.restore(
            self.directory, item=state_shape, restore_args=restore_args
        )

        test_utils.assert_tree_equal(self, state.params, restored.params)
        test_utils.assert_tree_equal(self, state.opt_state, restored.opt_state)

    @mock.patch.object(atomicity.AtomicRenameTemporaryPath, 'finalize')
    def test_save_preempted(self, _):
      """Simulate effects of preemption."""
      with self.checkpointer(PyTreeCheckpointHandler()) as checkpointer:
        checkpointer.save(self.directory, self.pytree)
        self.wait_if_async(checkpointer)
        paths = list(self.directory.parent.iterdir())
        self.assertLen(paths, 1)
        tmp_dir = paths[0]
        self.assertFalse(utils.is_checkpoint_finalized(tmp_dir))
        self.assertTrue(utils.is_tmp_checkpoint(tmp_dir))
        with self.assertRaisesRegex(ValueError, 'Found incomplete checkpoint'):
          checkpointer.restore(tmp_dir)

    @mock.patch.object(step, 'is_gcs_path', autospec=True, return_value=True)
    def test_gcs(self, is_gcs_path):
      """Test normal operation in simulated GCS environment."""
      del is_gcs_path
      with self.checkpointer(PyTreeCheckpointHandler()) as checkpointer:
        path = self.directory / '0' / 'params'
        checkpointer.save(path, self.pytree)
        self.wait_if_async(checkpointer)
        restored = checkpointer.restore(
            path, restore_args=self.pytree_restore_args
        )
        test_utils.assert_tree_equal(self, self.pytree, restored)
        self.assertTrue((path / step._COMMIT_SUCCESS_FILE).exists())

    @mock.patch.object(step, 'is_gcs_path', autospec=True, return_value=True)
    @mock.patch.object(atomicity.CommitFileTemporaryPath, 'finalize')
    def test_save_preempted_gcs(self, *mock_args):
      """Simulate effects of preemption."""
      del mock_args
      with self.checkpointer(PyTreeCheckpointHandler()) as checkpointer:
        checkpointer.save(self.directory, self.pytree)
        self.wait_if_async(checkpointer)
        paths = list(self.directory.parent.iterdir())
        self.assertLen(paths, 1)
        tmp_dir = paths[0]
        self.assertFalse(utils.is_checkpoint_finalized(tmp_dir))
        self.assertTrue(utils.is_tmp_checkpoint(tmp_dir))
        with self.assertRaisesRegex(ValueError, 'Found incomplete checkpoint'):
          checkpointer.restore(tmp_dir)
        self.assertFalse((tmp_dir / step._COMMIT_SUCCESS_FILE).exists())

    def test_configure_atomicity(self):
      """Test case."""
      with self.checkpointer(
          PyTreeCheckpointHandler(),
          temporary_path_class=atomicity.CommitFileTemporaryPath,
      ) as checkpointer:
        path = self.directory / '0' / 'params'
        checkpointer.save(path, self.pytree)
        self.wait_if_async(checkpointer)
        restored = checkpointer.restore(
            path, restore_args=self.pytree_restore_args
        )
        test_utils.assert_tree_equal(self, self.pytree, restored)
        self.assertTrue((path / step._COMMIT_SUCCESS_FILE).exists())  # pylint: disable=protected-access

    def test_legacy_api(self):
      """Test case."""

      # Skip CheckpointArgs registration.
      class MyCheckpointHandler(
          async_checkpoint_handler.AsyncCheckpointHandler
      ):

        async def async_save(self, directory, item, extra_args):
          return []

        def save(self, directory, item, extra_args):
          asyncio_utils.run_sync(self.async_save(directory, item, extra_args))

        def restore(self, directory, item=None, extra_args=None):
          return f'restored {item} with {extra_args} from {directory}.'

      with self.checkpointer(MyCheckpointHandler()) as ckptr:
        item = {'a': 1, 'b': 2}
        extra_args = {'special_args': True}
        ckptr.save(self.directory, item, extra_args)
        self.wait_if_async(ckptr)
        restored = ckptr.restore(
            self.directory, item=item, extra_args=extra_args
        )
        self.assertEqual(
            restored,
            f'restored {item} with {extra_args} from {self.directory}.',
        )

    def test_composite_handler(self):
      """Test case."""

      with self.checkpointer(
          composite_checkpoint_handler.CompositeCheckpointHandler()
      ) as ckptr:
        ckptr.save(
            self.directory,
            args.Composite(state=args.PyTreeSave(self.pytree)),
        )
        self.wait_if_async(ckptr)
        restored = ckptr.restore(
            self.directory,
            args=args.Composite(
                state=args.PyTreeRestore(restore_args=self.pytree_restore_args)
            ),
        )
        test_utils.assert_tree_equal(self, self.pytree, restored.state)

    @parameterized.parameters((5,), (9,))
    def test_memory_limiting(self, limit_bytes):
      """Test case."""
      sleep_time = 1.0
      handler, tree, restore_args = test_utils.concurrent_gb_test_setup()
      checkpointer = self.checkpointer(handler)

      byte_limiter = test_utils.get_byte_limiter(limit_bytes, sleep_time)
      with mock.patch.object(
          serialization,
          'get_byte_limiter',
          new=lambda _: byte_limiter,
      ):
        checkpointer.save(self.directory, tree)
      self.wait_if_async(checkpointer)
      completion_times = byte_limiter.completion_times
      # Replicated shards are handled within the _write_array_shard function.
      # Since shards are only saved once per replica, we only have to check
      # the primary process.
      if multihost.process_index() == 0:
        self.assertLen(completion_times, len(jax.tree.leaves(tree)))
        test_utils.assert_every_n_is_x_apart(
            self,
            completion_times,
            limit_bytes // np.int32().itemsize,
            sleep_time,
        )

      byte_limiter = test_utils.get_byte_limiter(limit_bytes, sleep_time)
      with mock.patch.object(
          serialization,
          'get_byte_limiter',
          new=lambda _: byte_limiter,
      ):
        checkpointer.restore(self.directory, restore_args=restore_args)
      completion_times = byte_limiter.completion_times
      self.assertLen(
          completion_times,
          len(jax.tree.leaves(tree)),
      )
      test_utils.assert_every_n_is_x_apart(
          self,
          completion_times,
          limit_bytes // np.int32().itemsize,
          sleep_time,
      )

    @parameterized.named_parameters(
        dict(
            testcase_name='default',
            save_input_provider=test_tree_utils.NamedTupleWithNestedAttributes,
            expected_restored_provider=(
                lambda: dict(
                    nested_mu_nu=None,
                    nested_dict=None,
                    nested_tuple=None,
                    nested_empty_named_tuple=None,
                    my_empty_chex=None,
                )
            ),
        ),
        dict(
            testcase_name='full',
            save_input_provider=(
                lambda: test_tree_utils.NamedTupleWithNestedAttributes(
                    nested_mu_nu=test_tree_utils.MuNu(
                        mu=test_utils.setup_sharded_pytree()[0],
                        nu=np.asarray([1, 2, 3]),
                    ),
                    nested_dict=dict(a=test_utils.setup_sharded_pytree()[0]),
                    nested_tuple=(
                        test_utils.setup_sharded_pytree()[0],
                        test_utils.setup_sharded_pytree()[0],
                    ),
                    nested_empty_named_tuple=test_tree_utils.EmptyNamedTuple(),
                    my_empty_chex=test_tree_utils.MyEmptyChex(),
                )
            ),
            expected_restored_provider=(
                lambda: dict(
                    nested_mu_nu=dict(
                        mu=test_utils.setup_sharded_pytree()[0],
                        nu=np.asarray([1, 2, 3]),
                    ),
                    nested_dict=dict(a=test_utils.setup_sharded_pytree()[0]),
                    nested_tuple=[
                        test_utils.setup_sharded_pytree()[0],
                        test_utils.setup_sharded_pytree()[0],
                    ],  # b/365169723
                    nested_empty_named_tuple={},
                    my_empty_chex={},
                )
            ),
        ),
        dict(
            testcase_name='partial',
            save_input_provider=(
                lambda: test_tree_utils.NamedTupleWithNestedAttributes(
                    nested_mu_nu=test_tree_utils.MuNu(
                        mu=None,
                        nu=np.asarray([1, 2, 3]),
                    ),
                    nested_dict=None,
                    nested_tuple=None,
                    nested_empty_named_tuple=test_tree_utils.EmptyNamedTuple(),
                )
            ),
            expected_restored_provider=(
                lambda: dict(
                    nested_mu_nu=dict(
                        mu=None,
                        nu=np.asarray([1, 2, 3]),
                    ),
                    nested_dict=None,
                    nested_tuple=None,
                    nested_empty_named_tuple={},
                    my_empty_chex=None,
                )
            ),
        ),
    )
    def test_save_restore_named_tuple(
        self,
        save_input_provider: Callable[[], PyTree],
        expected_restored_provider: Callable[[], PyTree],
    ):
      """Basic save and restore test."""
      with self.subTest('legacy_metadata'):
        with self.checkpointer(
            PyTreeCheckpointHandler(
                pytree_metadata_options=tree_metadata.PyTreeMetadataOptions(
                    support_rich_types=False
                )
            )
        ) as checkpointer:
          directory = self.directory / 'legacy_metadata'
          checkpointer.save(directory, save_input_provider())
          self.wait_if_async(checkpointer)
          restored = checkpointer.restore(directory)
          test_utils.assert_tree_equal(
              self, restored, expected_restored_provider()
          )

      with self.subTest('rich_typed_metadata'):
        with self.checkpointer(
            PyTreeCheckpointHandler(
                pytree_metadata_options=tree_metadata.PyTreeMetadataOptions(
                    support_rich_types=True
                )
            )
        ) as checkpointer:
          directory = self.directory / 'rich_typed_metadata'
          checkpointer.save(directory, save_input_provider())
          self.wait_if_async(checkpointer)
          # TODO: b/365169723 - Update this test when restore is ready.
          with self.assertRaises(NotImplementedError):
            _ = checkpointer.restore(directory)

    def test_save_step_metadata(self):
      """Basic save and restore test."""
      with self.checkpointer(PyTreeCheckpointHandler()) as checkpointer:
        directory = self.directory / 'step_metadata'
        checkpointer.save(
            directory,
            args.PyTreeSave(self.pytree),
            custom_metadata={'a': 1, 'b': 2}
        )
        self.wait_if_async(checkpointer)

        serialized_metadata = checkpoint.metadata_store(
            enable_write=False
        ).read(checkpoint.step_metadata_file_path(directory))
        self.assertIsNotNone(serialized_metadata)
        step_metadata = step_metadata_serialization.deserialize(
            serialized_metadata
        )

        self.assertEqual(
            step_metadata.item_handlers, checkpointer._handler.typestr()
        )
        self.assertIsNone(step_metadata.item_metadata)
        self.assertEqual(step_metadata.metrics, {})
        self.assertEqual(
            step_metadata.performance_metrics,
            step_statistics.SaveStepStatistics(),
        )
        self.assertGreater(step_metadata.init_timestamp_nsecs, 0)
        self.assertGreater(step_metadata.commit_timestamp_nsecs, 0)
        self.assertEqual(step_metadata.custom_metadata, {'a': 1, 'b': 2})

    async def test_save_with_failing_array_metadata_store_in_finalize(self):
      """ArrayMetadata validation in CheckpointHandler.finalize()."""

      class IncompleteArrayMetadataSerDeserializer(
          array_metadata_store_lib.SerDeserializer
      ):
        """process 0 metadata should be complete, others should be incomplete.

        Valid data in process 0 will enable successful pytree metadata write.
        Incomplete data in other processes will fail ArrayMetadata validation in
        CheckpointHandler.finalize().
        """

        def serialize(
            self, array_metadatas: Sequence[array_metadata.ArrayMetadata]
        ) -> str:
          if multihost.process_index() == 0:
            return super().serialize(array_metadatas)
          else:
            return super().serialize([array_metadatas[0]])  # incomplete

      array_metadata_store = array_metadata_store_lib.Store(
          ser_deser=IncompleteArrayMetadataSerDeserializer()
      )
      type_handler_registry = copy.deepcopy(
          type_handlers.GLOBAL_TYPE_HANDLER_REGISTRY
      )
      array_handler = type_handlers.ArrayHandler(
          array_metadata_store=array_metadata_store
      )
      type_handler_registry.add(jax.Array, array_handler, override=True)

      class InterceptingValidator(array_metadata_store_lib.Validator):
        """Intercepts ArrayMetadata validation in CheckpointHandler.finalize()."""

        def __init__(self):
          self.error = None

        def validate_all_array_metadatas(self, array_metadatas):
          try:
            super().validate_all_array_metadatas(array_metadatas)
          except ValueError as e:
            self.error = e

        def maybe_raise_error(self):
          if self.error is not None:
            raise self.error

      array_metadata_validator = InterceptingValidator()
      with self.checkpointer(
          PyTreeCheckpointHandler(
              type_handler_registry=type_handler_registry,
              array_metadata_validator=array_metadata_validator,
          )
      ) as checkpointer:
        checkpointer.save(self.directory, self.pytree)

      if multihost.process_index() == 0:
        with self.assertRaisesRegex(
            ValueError,
            'ArrayMetadata Store contains different number of params',
        ):
          array_metadata_validator.maybe_raise_error()
      self.assertIsNotNone(await array_metadata_store.read(self.directory))

    def test_partial_restore_with_placeholder(self):
      """Basic save and restore test."""
      with self.checkpointer(PyTreeCheckpointHandler()) as save_checkpointer:
        directory = self.directory / 'partial_restore'
        save_checkpointer.save(
            directory,
            args.PyTreeSave(self.pytree),
        )

      with self.subTest('success'):
        reference_item = self.empty_pytree.copy()
        reference_item['b'] = PLACEHOLDER
        reference_item['c']['e'] = PLACEHOLDER

        expected = self.pytree.copy()
        expected['b'] = PLACEHOLDER
        expected['c']['e'] = PLACEHOLDER

        with self.checkpointer(
            PyTreeCheckpointHandler()
        ) as restore_checkpointer:
          restored = restore_checkpointer.restore(
              directory,
              args=pytree_checkpoint_handler.PyTreeRestoreArgs(
                  item=reference_item,
                  restore_args=self.pytree_restore_args,
              ),
          )
          test_utils.assert_tree_equal(self, expected, restored)

      with self.subTest('missing_leaf'):
        reference_item = self.empty_pytree.copy()
        reference_item['b'] = PLACEHOLDER
        reference_item['c']['e'] = PLACEHOLDER
        del reference_item['c']['a']

        with self.checkpointer(
            PyTreeCheckpointHandler()
        ) as restore_checkpointer:
          with self.assertRaisesRegex(
              ValueError, 'User-provided restore item and on-disk value'
          ):
            restore_checkpointer.restore(
                directory,
                args=pytree_checkpoint_handler.PyTreeRestoreArgs(
                    item=reference_item,
                    restore_args=self.pytree_restore_args,
                ),
            )

      with self.subTest('non_leaf_placeholder'):
        reference_item = self.empty_pytree.copy()
        reference_item['c'] = PLACEHOLDER

        with self.checkpointer(
            PyTreeCheckpointHandler()
        ) as restore_checkpointer:
          with self.assertRaisesRegex(
              ValueError, 'User-provided restore item and on-disk value'
          ):
            restore_checkpointer.restore(
                directory,
                args=pytree_checkpoint_handler.PyTreeRestoreArgs(
                    item=reference_item,
                    restore_args=self.pytree_restore_args,
                ),
            )

    def test_partial_restore_with_omission(self):
      """Basic save and restore test."""
      with self.checkpointer(PyTreeCheckpointHandler()) as save_checkpointer:
        directory = self.directory / 'partial_restore'
        save_checkpointer.save(
            directory,
            args.PyTreeSave(self.pytree),
        )

      with self.subTest('success'):
        with self.checkpointer(
            PyTreeCheckpointHandler()
        ) as restore_checkpointer:
          reference_item = {
              'a': 0,
              'c': {
                  'a': 0,
              },
          }
          expected = {
              'a': self.pytree['a'],
              'c': {
                  'a': self.pytree['c']['a'],
              },
          }
          restored = restore_checkpointer.restore(
              directory,
              args=pytree_checkpoint_handler.PyTreeRestoreArgs(
                  item=reference_item,
                  restore_args=self.pytree_restore_args,
                  partial_restore=True,
              ),
          )
          test_utils.assert_tree_equal(self, expected, restored)

      with self.subTest('extra_leaf'):
        with self.checkpointer(
            PyTreeCheckpointHandler()
        ) as restore_checkpointer:
          reference_item = {
              'a': 0,
              'c': {
                  'a': 0,
              },
              'z': 0,
          }
          with self.assertRaisesRegex(
              ValueError,
              r"Missing 1 keys in structure path \(\), including: \['z'\]",
          ):
            restore_checkpointer.restore(
                directory,
                args=pytree_checkpoint_handler.PyTreeRestoreArgs(
                    item=reference_item,
                    restore_args=self.pytree_restore_args,
                    partial_restore=True,
                ),
            )



# Copyright 2025 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Base class for v0/v1 compatibility checkpointer tests."""

from typing import Any

from absl.testing import parameterized
from orbax.checkpoint import args as v0_args
from orbax.checkpoint import checkpoint_manager
from orbax.checkpoint import handlers as v0_handlers
import orbax.checkpoint.experimental.v1 as ocp
from orbax.checkpoint.experimental.v1._src.testing import handler_utils
from orbax.checkpoint.experimental.v1._src.training import checkpointer_test_base
from orbax.checkpoint.experimental.v1._src.tree import types as tree_types


Checkpointer = ocp.training.Checkpointer
RootMetadata = ocp.training.RootMetadata
Foo = handler_utils.Foo
Bar = handler_utils.Bar
Baz = handler_utils.Baz


def build_handler_registry_and_checkpoint_args(
    checkpointables: dict[str, Any],
) -> tuple[v0_handlers.CheckpointHandlerRegistry, v0_args.Composite]:
  """Builds v0 handler registry and checkpoint args from checkpointables."""
  checkpoint_args = {}
  handler_registry = {}
  for name, checkpointable in checkpointables.items():
    if name == 'pytree':
      checkpoint_args['pytree'] = v0_args.PyTreeSave(checkpointable)
      handler_registry['pytree'] = v0_handlers.PyTreeCheckpointHandler()
    elif isinstance(checkpointable, (Foo, Bar, Baz)):
      checkpoint_args[name] = handler_utils.DataclassSaveArgs(checkpointable)
      handler_registry[name] = handler_utils.DataclassCheckpointHandler()
    else:
      raise NotImplementedError(
          f'Unsupported checkpointable: name={name}, value={checkpointable}'
      )
  return (
      v0_handlers.create_default_handler_registry(**handler_registry),
      v0_args.Composite(**checkpoint_args),
  )


class V0v1CompatibilityCheckpointerTestBase(
    checkpointer_test_base.CheckpointerTestBase.Test
):
  """Base class for v0/v1 compatibility checkpointer tests."""

  def create_checkpoint_manager(
      self,
      checkpointer: Checkpointer,
      handler_registry: v0_handlers.CheckpointHandlerRegistry,
  ) -> checkpoint_manager.CheckpointManager:
    return checkpoint_manager.CheckpointManager(
        self.directory,
        options=checkpoint_manager.CheckpointManagerOptions(
            create=False,
            enable_async_checkpointing=False,
            save_decision_policy=(
                checkpointer._manager._options.save_decision_policy  # pylint: disable=protected-access
            ),
            step_name_format=checkpointer._manager._options.step_name_format,  # pylint: disable=protected-access
            prevent_write_metrics=False,  # Write metrics on disk.
            best_fn=lambda metrics: metrics['loss'],  # Write metrics on disk.
        ),
        handler_registry=handler_registry,
    )

  def save_pytree(
      self,
      checkpointer: Checkpointer,
      step: int,
      pytree: tree_types.PyTreeOf[tree_types.LeafType],
      metrics: tree_types.JsonType | None = None,
      custom_metadata: tree_types.JsonType | None = None,
  ) -> bool:
    """Saves a pytree checkpoint with v0 CheckpointManager."""
    self.skipTest('b/422287659')
    manager = self.create_checkpoint_manager(
        checkpointer,
        handler_registry=v0_handlers.create_default_handler_registry(
            pytree=v0_handlers.PyTreeCheckpointHandler()
        ),
    )
    with manager:
      result = manager.save(
          step,
          args=v0_args.Composite(pytree=v0_args.PyTreeSave(pytree)),
          metrics=metrics,
          custom_metadata=custom_metadata,
      )
      manager.wait_until_finished()
      checkpointer.reload()  # Reload v1 checkpointer to sync with v0 ckpts.
      return result

  def save_checkpointables(
      self,
      checkpointer: Checkpointer,
      step: int,
      checkpointables: dict[str, Any],
      metrics: tree_types.JsonType | None = None,
      custom_metadata: tree_types.JsonType | None = None,
  ) -> bool:
    handler_registry, checkpoint_args = (
        build_handler_registry_and_checkpoint_args(checkpointables)
    )
    manager = self.create_checkpoint_manager(checkpointer, handler_registry)
    with manager:
      result = manager.save(
          step,
          args=checkpoint_args,
          metrics=metrics,
          custom_metadata=custom_metadata,
      )
      manager.wait_until_finished()
      checkpointer.reload()  # Reload v1 checkpointer to sync with v0 ckpts.
      return result

  def test_skips_when_ongoing_save(self):
    self.skipTest('Not relevant for v0/v1 compatibility.')

  def test_save_pytree_async(self):
    self.skipTest('Not relevant for v0/v1 compatibility.')

  def test_close(self):
    self.skipTest('Not relevant for v0/v1 compatibility.')

  def test_context_manager_close(self):
    self.skipTest('Not relevant for v0/v1 compatibility.')

  @parameterized.product(
      reinitialize_checkpointer=(True, False),
  )
  def test_root_metadata(self, reinitialize_checkpointer):
    if not reinitialize_checkpointer:
      self.skipTest('Not relevant for v0/v1 compatibility.')

    with checkpoint_manager.CheckpointManager(
        self.directory,
        options=checkpoint_manager.CheckpointManagerOptions(
            create=False,
            enable_async_checkpointing=False,
        ),
        handler_registry=v0_handlers.create_default_handler_registry(
            pytree=v0_handlers.PyTreeCheckpointHandler()
        ),
        metadata={'foo': 'bar'},  # saved during manager instance creation.
    ):
      pass
    # Does not overwrite custom metadata.
    checkpointer = Checkpointer(self.directory, custom_metadata={'baz': 2})
    self.enter_context(checkpointer)
    root_metadata = checkpointer.root_metadata()
    self.assertIsInstance(root_metadata, RootMetadata)
    self.assertDictEqual(root_metadata.custom_metadata, {'foo': 'bar'})

  def test_custom_checkpointables(self):
    # Use named handler to override v0 checkpoint_handlers.
    checkpointables_options = (
        ocp.options.CheckpointablesOptions.create_with_handlers(
            foo=handler_utils.FooHandler,
            bar=handler_utils.BarHandler,
        )
    )
    self.enter_context(
        ocp.Context(checkpointables_options=checkpointables_options)
    )
    super().test_custom_checkpointables()

  def test_load_with_switched_abstract_checkpointables(self):
    # Use named handler to override v0 checkpoint_handlers.
    checkpointables_options = (
        ocp.options.CheckpointablesOptions.create_with_handlers(
            bar=handler_utils.FooHandler,
            foo=handler_utils.BarHandler,
        )
    )
    self.enter_context(
        ocp.Context(checkpointables_options=checkpointables_options)
    )
    super().test_load_with_switched_abstract_checkpointables()

  def test_different_custom_checkpointables(self):
    # Use named handler to override v0 checkpoint_handlers.
    checkpointables_options = (
        ocp.options.CheckpointablesOptions.create_with_handlers(
            foo=handler_utils.FooHandler,
            bar=handler_utils.BarHandler,
        )
    )
    self.enter_context(
        ocp.Context(checkpointables_options=checkpointables_options)
    )
    super().test_different_custom_checkpointables()

  def test_custom_save_decision_policy(self):
    self.skipTest('b/422287659')

  def test_overwrites(self):
    self.skipTest('b/422287659')

  def test_steps(self):
    self.skipTest('b/422287659')

  def test_step_already_exists(self):
    self.skipTest('b/422287659')



# Copyright 2025 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Base class for v0/v1 compatibility save/load tests."""

# pylint: disable=missing-class-docstring,protected-access,missing-function-docstring

from __future__ import annotations

from absl.testing import parameterized
from etils import epath
from orbax.checkpoint import args as args_lib
from orbax.checkpoint import test_utils
from orbax.checkpoint._src.checkpointers import checkpointer as v0_checkpointer
from orbax.checkpoint._src.checkpointers import standard_checkpointer
from orbax.checkpoint._src.handlers import composite_checkpoint_handler
import orbax.checkpoint.experimental.v1 as ocp
from orbax.checkpoint.experimental.v1._src.path import types as path_types
from orbax.checkpoint.experimental.v1._src.synchronization import multihost
from orbax.checkpoint.experimental.v1._src.testing import array_utils as array_test_utils
from orbax.checkpoint.experimental.v1._src.tree import types as tree_types


PyTree = tree_types.PyTree
Path = path_types.Path

create_sharded_pytree = array_test_utils.create_sharded_pytree


class CompatibilitySaveLoadTestBase:

  class Test(parameterized.TestCase):

    def setUp(self):
      super().setUp()

      self.root_directory = epath.Path(
          self.create_tempdir(name='root').full_path
      )
      self.ckpt_directory = (
          epath.Path(self.create_tempdir(name='direct').full_path) / 'ckpt'
      )
      self.pytree, self.abstract_pytree = create_sharded_pytree()

      test_utils.set_tensorstore_driver_for_test()
      test_utils.sync_global_processes('CompatibilityTest:setup_complete')

    def tearDown(self):
      super().tearDown()
      test_utils.sync_global_processes('CompatibilityTest:teardown_complete')

    def save_v0_checkpoint(self, directory: Path):
      with standard_checkpointer.StandardCheckpointer() as checkpointer:
        checkpointer.save(directory, self.pytree)

    def save_v0_checkpoints(
        self, base_dir: Path, *, checkpointable_names: list[str]
    ):
      args = args_lib.Composite(**{
          checkpointable_name: args_lib.StandardSave(self.pytree)
          for checkpointable_name in checkpointable_names
      })
      with v0_checkpointer.Checkpointer(
          composite_checkpoint_handler.CompositeCheckpointHandler()
      ) as checkpointer:
        checkpointer.save(base_dir, args)

    def test_async_load(self):
      with self.assertRaises(NotImplementedError):
        ocp.load_pytree_async(
            self.root_directory,
        )
      with self.assertRaises(NotImplementedError):
        ocp.load_checkpointables_async(
            self.root_directory,
        )

    @parameterized.product(with_abstract_pytree=[True, False])
    def test_load_v0_checkpoint_with_v1_load_pytree(
        self,
        with_abstract_pytree: bool,
    ):

      checkpointable_names = ['default', 'state', 'pytree']
      step_dir = self.root_directory / 'load_pytree_0'
      self.save_v0_checkpoints(
          step_dir,
          checkpointable_names=checkpointable_names,
      )
      self.save_v0_checkpoint(self.ckpt_directory)

      with self.subTest('no_checkpointable_name'):
        loaded = ocp.load_pytree(
            step_dir,
            self.abstract_pytree if with_abstract_pytree else None,
        )
        test_utils.assert_tree_equal(self, self.pytree, loaded)

      with self.subTest('no_checkpointable_name_error'):
        with self.assertRaisesRegex(
            FileNotFoundError,
            'must contain a subdirectory named "pytree"',
        ):
          ocp.load_pytree(
              self.ckpt_directory,
              self.abstract_pytree if with_abstract_pytree else None,
          )
        with self.assertRaisesRegex(
            FileNotFoundError,
            'must contain a subdirectory named "pytree"',
        ):
          ocp.load_pytree(
              self.root_directory,
              self.abstract_pytree if with_abstract_pytree else None,
          )

      for checkpointable_name in checkpointable_names:
        with self.subTest(f'pass_{checkpointable_name}'):
          loaded = ocp.load_pytree(
              step_dir,
              self.abstract_pytree if with_abstract_pytree else None,
              checkpointable_name=checkpointable_name,
          )
          test_utils.assert_tree_equal(self, self.pytree, loaded)

        with self.subTest(f'pass_{checkpointable_name}_error'):
          with self.assertRaisesRegex(
              FileNotFoundError,
              f'must contain a subdirectory named "{checkpointable_name}"',
          ):
            ocp.load_pytree(
                self.ckpt_directory,
                self.abstract_pytree if with_abstract_pytree else None,
                checkpointable_name=checkpointable_name,
            )
          with self.assertRaisesRegex(
              FileNotFoundError,
              f'must contain a subdirectory named "{checkpointable_name}"',
          ):
            ocp.load_pytree(
                self.root_directory,
                self.abstract_pytree if with_abstract_pytree else None,
                checkpointable_name=checkpointable_name,
            )

      with self.subTest('pass_none'):
        loaded = ocp.load_pytree(
            self.ckpt_directory,
            self.abstract_pytree if with_abstract_pytree else None,
            checkpointable_name=None,
        )
        test_utils.assert_tree_equal(self, self.pytree, loaded)

      with self.subTest('pass_none_error'):
        with self.assertRaisesRegex(
            FileNotFoundError, 'does not contain a PyTree metadata file'
        ):
          ocp.load_pytree(
              step_dir,
              self.abstract_pytree if with_abstract_pytree else None,
              checkpointable_name=None,
          )
        with self.assertRaisesRegex(
            FileNotFoundError, 'does not contain a PyTree metadata file'
        ):
          ocp.load_pytree(
              self.root_directory,
              self.abstract_pytree if with_abstract_pytree else None,
              checkpointable_name=None,
          )

    @parameterized.product(
        checkpointable_name=['default', 'state'],
        with_abstract_pytree=[True, False],
    )
    def test_load_v0_checkpoint_with_v1_load_checkpointables(
        self,
        checkpointable_name: str,
        with_abstract_pytree: bool,
    ):

      checkpointable_names = ['default', 'state']
      step_dir = self.root_directory / 'load_checkpointables_0'
      self.save_v0_checkpoints(
          step_dir,
          checkpointable_names=checkpointable_names,
      )
      self.save_v0_checkpoint(self.ckpt_directory)

      abstract_checkpointables = (
          {checkpointable_name: self.abstract_pytree}
          if with_abstract_pytree
          else None
      )

      with self.subTest('with_context'):
        checkpointables_options = (
            ocp.options.CheckpointablesOptions.create_with_handlers(
                **{checkpointable_name: ocp.handlers.PyTreeHandler}
            )
        )
        with ocp.Context(checkpointables_options=checkpointables_options):
          loaded = ocp.load_checkpointables(step_dir, abstract_checkpointables)
          test_utils.assert_tree_equal(
              self, self.pytree, loaded[checkpointable_name]
          )

      with self.subTest('without_context'):
        loaded = ocp.load_checkpointables(step_dir, abstract_checkpointables)
        test_utils.assert_tree_equal(
            self, self.pytree, loaded[checkpointable_name]
        )

      with self.subTest('error_with_checkpoint_path'):
        with self.assertRaisesRegex(
            ValueError,
            'If you intended to load a pytree checkpoint from the given path',
        ):
          ocp.load_checkpointables(
              self.ckpt_directory, abstract_checkpointables
          )
      with self.subTest('error_with_root_path'):
        with self.assertRaisesRegex(
            FileNotFoundError,
            'Please ensure the path specified for loading points to a valid'
            ' Orbax checkpoint',
        ):
          ocp.load_checkpointables(
              self.root_directory, abstract_checkpointables
          )



# Copyright 2025 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Additional functions for training.py, which trains a simple CNN model with MNIST data and demonstrates how to use orbax checkpointing."""

from flax import nnx
import grain.python as pygrain
import jax
import jax.numpy as jnp
import numpy as np
import tensorflow_datasets as tfds

PartitionSpec = jax.sharding.PartitionSpec


class DotReluDot(nnx.Module):
  """A simple model with two linear layers and a ReLU activation."""

  def __init__(self, depth: int, rngs: nnx.Rngs):
    init_fn = nnx.initializers.lecun_normal()

    self.dot1 = nnx.Linear(
        in_features=784,
        out_features=depth,
        kernel_init=nnx.with_partitioning(init_fn, (None, 'model')),
        use_bias=False,
        rngs=rngs,
    )

    self.w2 = nnx.Param(
        init_fn(rngs.params(), (depth, depth)),
        sharding=('model', None),
    )

  def __call__(self, x: jax.Array) -> jax.Array:
    x = x.reshape((x.shape[0], -1))
    y = self.dot1(x)
    y = jax.nn.relu(y)
    y = jax.lax.with_sharding_constraint(
        y,
        PartitionSpec(
            'data',
        ),
    )
    z = jnp.dot(y, self.w2.value)
    return z


def process_sample(sample: dict[str, jax.Array]) -> dict[str, jax.Array]:
  """Converts the image to float32 and normalizes it."""
  return {
      'image': sample['image'].astype(np.float32) / 255.0,
      'label': sample['label'],
  }


def create_dataset(split: str, batch_size: int) -> pygrain.IterDataset:
  """Creates a Grain-based dataset for a given split."""
  """TODO(zachmeyers): parameterize seed."""  # pylint: disable=pointless-string-statement
  dataset = (
      pygrain.MapDataset.source(tfds.data_source('mnist', split=split))
      .map(process_sample)
      .seed(seed=45)
      .shuffle()
      .batch(batch_size, drop_remainder=True)
  )
  return dataset.to_iter_dataset()



# Copyright 2025 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Test base classes for emergency.CheckpointManager."""

# pylint: disable=protected-access

import functools
import json
from typing import Any, Optional
from unittest import mock

from absl import logging
from absl.testing import flagsaver
from absl.testing import parameterized
from etils import epath
import jax
import numpy as np
import optax
from orbax.checkpoint import args as args_lib
from orbax.checkpoint import test_utils
from orbax.checkpoint import utils
from orbax.checkpoint._src.handlers import pytree_checkpoint_handler
from orbax.checkpoint._src.multihost import multihost
from orbax.checkpoint._src.multihost import multislice
from orbax.checkpoint.experimental.emergency import checkpoint_manager
from orbax.checkpoint.experimental.emergency import mesh_consistency
from orbax.checkpoint.experimental.emergency import process_metadata_checkpoint_handler

PyTree = Any
PyTreeCheckpointHandler = pytree_checkpoint_handler.PyTreeCheckpointHandler
PyTreeSaveArgs = pytree_checkpoint_handler.PyTreeSaveArgs
PyTreeRestoreArgs = pytree_checkpoint_handler.PyTreeRestoreArgs

CheckpointManager = checkpoint_manager.CheckpointManager
LocalCheckpointManager = checkpoint_manager._LocalCheckpointManager  # pylint: disable=protected-access
CheckpointManagerOptions = checkpoint_manager.CheckpointManagerOptions
LocalCheckpointOptions = checkpoint_manager.LocalCheckpointOptions
PersistentCheckpointOptions = checkpoint_manager.PersistentCheckpointOptions
barrier_compatible_test = test_utils.barrier_compatible_test
assert_tree_equal = test_utils.assert_tree_equal
get_fake_global_mesh_for_slices = test_utils.get_fake_global_mesh_for_slices
_STATE_ITEM_NAME = 'state'


def swap_slices_in_mesh(
    mesh: jax.sharding.Mesh, *, replica_axis_index: int = 0
) -> jax.sharding.Mesh:
  """Reverses the ordering of devices such that slices swap IDs."""
  devices = []
  for slice_id in range(
      multislice.slice_count(mesh, replica_axis_index=replica_axis_index)
  ):
    devices.append(
        multislice.slice_devices(
            mesh, replica_id=slice_id, replica_axis_index=replica_axis_index
        )
    )
  devices.reverse()
  devices = np.stack(devices, axis=replica_axis_index)
  return jax.sharding.Mesh(devices, mesh.axis_names)


def _replace_abstract_array_sharding_with_mesh(
    sds: jax.ShapeDtypeStruct, mesh: jax.sharding.Mesh
) -> jax.ShapeDtypeStruct:
  assert isinstance(sds.sharding, jax.sharding.NamedSharding)
  return sds.update(
      sharding=jax.sharding.NamedSharding(mesh, sds.sharding.spec)
  )


def get_composite_save_args(
    pytree: PyTree,
) -> args_lib.Composite:
  return args_lib.Composite(state=PyTreeSaveArgs(pytree))


class LocalCheckpointManagerTestBase:
  """Test case."""

  @barrier_compatible_test
  class Test(parameterized.TestCase):
    """Test case."""

    def make_global_mesh(
        self, replica_axis_index: int = 0
    ) -> jax.sharding.Mesh:
      """Creates a global mesh for testing purposes.

      This method sets up a global mesh with 8 devices across 4 processes,
      where each process has 2 local devices. The mesh is constructed based
      on provided slice processes and a specified replica axis.

      Args:
        replica_axis_index: The index of the replica axis in the resulting mesh.
          Must be either 0 or 1. Defaults to 0.

      Returns:
        A `jax.sharding.Mesh` object representing the fake global mesh.

      Raises:
        ValueError: If `replica_axis_index` is not 0 or 1.
      """
      if replica_axis_index not in [0, 1]:
        raise ValueError(
            'replica_axis_index must be 0 or 1 for this test. Got:'
            f' {replica_axis_index}.'
        )
      self.assertEqual(jax.device_count(), 8)
      self.assertEqual(jax.process_count(), 4)
      self.assertEqual(jax.local_device_count(), 2)

      # setup global mesh info for 2-slice tests
      slice_processes = [{0, 1}, {2, 3}]
      mesh = test_utils.get_fake_global_mesh_for_slices(
          slice_processes, replica_axis_index
      )
      if replica_axis_index == 0:
        assert mesh.devices.shape == (2, 4), mesh.devices.shape
      if replica_axis_index == 1:
        assert mesh.devices.shape == (4, 2), mesh.devices.shape
      return mesh

    def setUp(self):
      super().setUp()
      self.enter_context(
          flagsaver.flagsaver(
              experimental_orbax_use_distributed_process_id=True
          )
      )
      if not multihost.is_runtime_to_distributed_ids_initialized():
        multihost.initialize_runtime_to_distributed_ids()
        multihost.initialize_distributed_to_device_ids()

      self.global_mesh = self.make_global_mesh()

      self._fn = lambda ty: issubclass(ty, jax.Array)

      pytree, mesh_tree, axes_tree = test_utils.setup_sharded_pytree()
      doubled_pytree = test_utils.apply_function(pytree, lambda x: x * 2)

      self.pytree = pytree
      self.doubled_pytree = doubled_pytree
      self.mesh_tree = mesh_tree
      self.axes_tree = axes_tree

      # make sure each process is working on different directories
      self.local_directory = epath.Path(
          self.create_tempdir(
              name=f'checkpointing_test_pid{multihost.process_index()}'
          ).full_path
      )
      logging.info(
          'self.directory=%s',
          self.local_directory,
      )
      test_utils.set_tensorstore_driver_for_test()

      test_utils.sync_global_processes(
          'LocalCheckpointManagerTest:setup_complete'
      )

    def tearDown(self):
      super().tearDown()
      test_utils.sync_global_processes(
          'LocalCheckpointManagerTest:teardown_complete'
      )

    def test_local_save_restore(self):
      """Test case."""
      options = CheckpointManagerOptions(
          local=LocalCheckpointOptions(
              save_interval_steps=1,
              max_to_keep=1,
              debug_use_full_global_mesh=True,
          ),
          persistent=PersistentCheckpointOptions(
              save_interval_steps=1, max_to_keep=1
          ),
      )
      manager = LocalCheckpointManager(
          directory=self.local_directory,
          global_mesh=self.global_mesh,
          options=options,
      )

      manager.save(
          0,
          args=args_lib.Composite(
              state=PyTreeSaveArgs(self.pytree),
              process_metadata=process_metadata_checkpoint_handler.ProcessMetadataSaveArgs(
                  global_mesh=self.global_mesh
              ),
          ),
      )

      manager.wait_until_finished()

      restored = manager.restore(
          0,
          args=args_lib.Composite(
              state=PyTreeRestoreArgs(),
              process_metadata=process_metadata_checkpoint_handler.ProcessMetadataRestoreArgs(),
          ),
      )
      test_utils.assert_tree_equal(self, self.pytree, restored.state)

    def test_local_auto_cleanup(self):
      """Test case."""
      options = CheckpointManagerOptions(
          local=LocalCheckpointOptions(
              save_interval_steps=1,
              max_to_keep=1,
              debug_use_full_global_mesh=True,
          ),
          persistent=PersistentCheckpointOptions(
              save_interval_steps=1, max_to_keep=1
          ),
      )
      manager = LocalCheckpointManager(
          directory=self.local_directory,
          global_mesh=self.global_mesh,
          options=options,
      )

      manager.save(
          0,
          args=args_lib.Composite(
              state=PyTreeSaveArgs(self.pytree),
              process_metadata=process_metadata_checkpoint_handler.ProcessMetadataSaveArgs(
                  global_mesh=self.global_mesh
              ),
          ),
      )
      manager.wait_until_finished()

      manager.save(
          1,
          args=args_lib.Composite(
              state=PyTreeSaveArgs(self.doubled_pytree),
              process_metadata=process_metadata_checkpoint_handler.ProcessMetadataSaveArgs(
                  global_mesh=self.global_mesh
              ),
          ),
      )
      manager.wait_until_finished()

      # step 0 should have been removed
      with self.assertRaisesRegex(ValueError, 'No step path found'):
        manager.restore(
            0,
            args=args_lib.Composite(
                state=PyTreeRestoreArgs(),
                process_metadata=process_metadata_checkpoint_handler.ProcessMetadataRestoreArgs(),
            ),
        )

      restored = manager.restore(
          1,
          args=args_lib.Composite(
              state=PyTreeRestoreArgs(),
              process_metadata=process_metadata_checkpoint_handler.ProcessMetadataRestoreArgs,
          ),
      )

      test_utils.assert_tree_equal(self, self.doubled_pytree, restored.state)

    def test_default_max_to_keep_is_1(self):
      """Test case."""
      # the default max_to_keep should be 1 because we want to ensure that by
      # default, up to 2 steps are in local storage at the same time.
      options = CheckpointManagerOptions(
          local=LocalCheckpointOptions(
              save_interval_steps=1,
              # max_to_keep unset to use default value
              debug_use_full_global_mesh=True,
          ),
      )
      manager = LocalCheckpointManager(
          directory=self.local_directory,
          global_mesh=self.global_mesh,
          options=options,
      )
      manager.save(
          0,
          args=args_lib.Composite(
              state=PyTreeSaveArgs(self.pytree),
              process_metadata=process_metadata_checkpoint_handler.ProcessMetadataSaveArgs(
                  global_mesh=self.global_mesh
              ),
          ),
      )
      manager.save(
          1,
          args=args_lib.Composite(
              state=PyTreeSaveArgs(self.doubled_pytree),
              process_metadata=process_metadata_checkpoint_handler.ProcessMetadataSaveArgs(
                  global_mesh=self.global_mesh
              ),
          ),
      )
      manager.wait_until_finished()

      self.assertEqual(manager.all_steps(), [1])
      self.assertFalse((self.local_directory / '0').exists())
      self.assertTrue((self.local_directory / '1' / 'state').exists())
      self.assertTrue(
          (self.local_directory / '1' / 'process_metadata').exists()
      )

    @parameterized.parameters(
        (False),
        (True,),
    )
    def test_local_startup_cleanup(self, enable_async_checkpointing: bool):
      """Test case."""
      options = CheckpointManagerOptions(
          local=LocalCheckpointOptions(
              save_interval_steps=1,
              max_to_keep=3,
              debug_use_full_global_mesh=True,
          ),
          enable_async_checkpointing=enable_async_checkpointing,
      )
      manager = LocalCheckpointManager(
          directory=self.local_directory,
          global_mesh=self.global_mesh,
          options=options,
      )

      manager.save(
          0,
          args=args_lib.Composite(
              state=PyTreeSaveArgs(self.pytree),
              process_metadata=process_metadata_checkpoint_handler.ProcessMetadataSaveArgs(
                  global_mesh=self.global_mesh
              ),
          ),
      )
      manager.save(
          1,
          args=args_lib.Composite(
              state=PyTreeSaveArgs(self.pytree),
              process_metadata=process_metadata_checkpoint_handler.ProcessMetadataSaveArgs(
                  global_mesh=self.global_mesh
              ),
          ),
      )
      manager.save(
          2,
          args=args_lib.Composite(
              state=PyTreeSaveArgs(self.doubled_pytree),
              process_metadata=process_metadata_checkpoint_handler.ProcessMetadataSaveArgs(
                  global_mesh=self.global_mesh
              ),
          ),
      )
      manager.wait_until_finished()

      # 3 steps should exist
      self.assertEqual(manager.all_steps(), [0, 1, 2])
      for i in range(3):
        self.assertTrue((self.local_directory / str(i)).exists())
        self.assertTrue(
            mesh_consistency.process_metadata_folder(
                self.local_directory / str(i) / 'process_metadata'
            ).exists()
        )

      # create a new manager with max_to_keep=1
      options = CheckpointManagerOptions(
          local=LocalCheckpointOptions(
              save_interval_steps=1,
              max_to_keep=1,
              debug_use_full_global_mesh=True,
          ),
          enable_async_checkpointing=enable_async_checkpointing,
      )
      manager2 = LocalCheckpointManager(
          directory=self.local_directory,
          global_mesh=self.global_mesh,
          options=options,
      )

      # manager2 should have gc'd step 0 and 1
      self.assertEqual(manager2.all_steps(), [2])
      self.assertFalse((self.local_directory / '0').exists())
      self.assertFalse((self.local_directory / '1').exists())
      self.assertTrue((self.local_directory / '2' / 'state').exists())
      self.assertTrue(
          (self.local_directory / '2' / 'process_metadata').exists()
      )

      # new step save should work
      manager2.save(
          3,
          args=args_lib.Composite(
              state=PyTreeSaveArgs(self.doubled_pytree),
              process_metadata=process_metadata_checkpoint_handler.ProcessMetadataSaveArgs(
                  global_mesh=self.global_mesh
              ),
          ),
      )
      manager2.wait_until_finished()
      self.assertEqual(manager2.all_steps(), [3])
      self.assertFalse((self.local_directory / '2').exists())
      self.assertTrue((self.local_directory / '3' / 'state').exists())
      self.assertTrue(
          (self.local_directory / '3' / 'process_metadata').exists()
      )

    @parameterized.parameters(
        (False),
        (True,),
    )
    def test_local_garbage_collection(self, enable_async_checkpointing: bool):
      """Test case."""
      options = CheckpointManagerOptions(
          local=LocalCheckpointOptions(
              save_interval_steps=1,
              max_to_keep=2,
              debug_use_full_global_mesh=True,
          ),
          persistent=PersistentCheckpointOptions(
              save_interval_steps=1, max_to_keep=1
          ),
          enable_async_checkpointing=enable_async_checkpointing,
      )
      manager = LocalCheckpointManager(
          directory=self.local_directory,
          global_mesh=self.global_mesh,
          options=options,
      )

      manager.save(
          0,
          args=args_lib.Composite(
              state=PyTreeSaveArgs(self.pytree),
              process_metadata=process_metadata_checkpoint_handler.ProcessMetadataSaveArgs(
                  global_mesh=self.global_mesh
              ),
          ),
      )
      manager.save(
          1,
          args=args_lib.Composite(
              state=PyTreeSaveArgs(self.pytree),
              process_metadata=process_metadata_checkpoint_handler.ProcessMetadataSaveArgs(
                  global_mesh=self.global_mesh
              ),
          ),
      )
      manager.save(
          2,
          args=args_lib.Composite(
              state=PyTreeSaveArgs(self.doubled_pytree),
              process_metadata=process_metadata_checkpoint_handler.ProcessMetadataSaveArgs(
                  global_mesh=self.global_mesh
              ),
          ),
      )
      manager.wait_until_finished()
      self.assertFalse((self.local_directory / '0').exists())
      self.assertTrue((self.local_directory / '1' / 'state').exists())
      self.assertTrue(
          (self.local_directory / '1' / 'process_metadata').exists()
      )

      self.assertTrue((self.local_directory / '2' / 'state').exists())
      self.assertTrue(
          (self.local_directory / '2' / 'process_metadata').exists()
      )

      if multihost.process_index() == 0:
        test_utils.empty_directory(self.local_directory)

      manager.reload()

      if multihost.process_index() == 0:
        self.assertFalse((self.local_directory / '1').exists())
        self.assertFalse((self.local_directory / '2').exists())
      else:
        self.assertTrue((self.local_directory / '1' / 'state').exists())
        self.assertTrue(
            (self.local_directory / '1' / 'process_metadata').exists()
        )
        self.assertTrue((self.local_directory / '2' / 'state').exists())
        self.assertTrue(
            (self.local_directory / '2' / 'process_metadata').exists()
        )

      # subsequent save should be successful and gc should work
      manager.save(
          3,
          args=args_lib.Composite(
              state=PyTreeSaveArgs(self.pytree),
              process_metadata=process_metadata_checkpoint_handler.ProcessMetadataSaveArgs(
                  global_mesh=self.global_mesh
              ),
          ),
      )
      manager.wait_until_finished()

      if multihost.process_index() == 0:
        self.assertFalse((self.local_directory / '2').exists())
        self.assertTrue((self.local_directory / '3' / 'state').exists())
        self.assertTrue(
            (self.local_directory / '3' / 'process_metadata').exists()
        )
      else:
        self.assertFalse((self.local_directory / '1').exists())
        self.assertTrue((self.local_directory / '2' / 'state').exists())
        self.assertTrue(
            (self.local_directory / '2' / 'process_metadata').exists()
        )
        self.assertTrue((self.local_directory / '3' / 'state').exists())
        self.assertTrue(
            (self.local_directory / '3' / 'process_metadata').exists()
        )

      manager.save(
          4,
          args=args_lib.Composite(
              state=PyTreeSaveArgs(self.doubled_pytree),
              process_metadata=process_metadata_checkpoint_handler.ProcessMetadataSaveArgs(
                  global_mesh=self.global_mesh
              ),
          ),
      )
      manager.save(
          5,
          args=args_lib.Composite(
              state=PyTreeSaveArgs(self.pytree),
              process_metadata=process_metadata_checkpoint_handler.ProcessMetadataSaveArgs(
                  global_mesh=self.global_mesh
              ),
          ),
      )
      manager.wait_until_finished()

      self.assertFalse((self.local_directory / '2').exists())
      self.assertFalse((self.local_directory / '3').exists())
      self.assertTrue((self.local_directory / '4' / 'state').exists())
      self.assertTrue(
          (self.local_directory / '4' / 'process_metadata').exists()
      )
      self.assertTrue((self.local_directory / '5' / 'state').exists())
      self.assertTrue(
          (self.local_directory / '5' / 'process_metadata').exists()
      )

    @parameterized.parameters(
        ([[], [], [], []], {0: {}, 1: {}}),
        ([[1], [1], [0], [0]], {0: {1}, 1: {0}}),
        ([[], [], [0], [0]], {0: {}, 1: {0}}),
        ([[], [], [0], []], {0: {}, 1: {}}),
        ([[], [], [0], [1]], {0: {}, 1: {}}),
        ([[], [0], [], [0]], {0: {}, 1: {}}),
        ([[0], [1], [0], [1]], {0: {}, 1: {0}}),
        ([[1, 2], [1, 2], [4], [4, 5]], {0: {1, 2}, 1: {4}}),
        ([[-1, 0], [-1, 0], [1, 2], [1, 2]], {0: {-1, 0}, 1: {1, 2}}),
        (
            [[-1, 0, 1], [-1, 0, 1], [0, 1, -1], [1, 0, -1]],
            {0: {-1, 0, 1}, 1: {-1, 0, 1}},
        ),
    )
    def test_common_steps_per_slice(self, process_steps, expectation):
      """Test case."""
      per_process_steps = {
          pid: steps for pid, steps in enumerate(process_steps)
      }
      result = checkpoint_manager._common_values_per_slice(  # pylint: disable=protected-access
          per_process_steps, self.global_mesh, replica_axis_index=0
      )
      self.assertSameElements(result, expectation)

    @parameterized.parameters((True, []), (True, [0]), (False, [0, 3]))
    def test_local_all_steps(self, steps_found, processes_to_clear):
      """Test case."""
      options = CheckpointManagerOptions(
          local=LocalCheckpointOptions(
              save_interval_steps=1,
              max_to_keep=4,
              debug_use_full_global_mesh=True,
          ),
      )
      manager = LocalCheckpointManager(
          directory=self.local_directory,
          global_mesh=self.global_mesh,
          options=options,
      )

      manager.save(
          0,
          args=args_lib.Composite(
              state=PyTreeSaveArgs(self.pytree),
              process_metadata=process_metadata_checkpoint_handler.ProcessMetadataSaveArgs(
                  global_mesh=self.global_mesh
              ),
          ),
      )
      manager.wait_until_finished()

      manager.save(
          1,
          args=args_lib.Composite(
              state=PyTreeSaveArgs(self.doubled_pytree),
              process_metadata=process_metadata_checkpoint_handler.ProcessMetadataSaveArgs(
                  global_mesh=self.global_mesh
              ),
          ),
      )
      manager.wait_until_finished()

      manager.save(
          2,
          args=args_lib.Composite(
              state=PyTreeSaveArgs(self.pytree),
              process_metadata=process_metadata_checkpoint_handler.ProcessMetadataSaveArgs(
                  global_mesh=self.global_mesh
              ),
          ),
      )
      manager.wait_until_finished()

      if multihost.process_index() in processes_to_clear:
        test_utils.empty_directory(self.local_directory)
      manager.reload()

      self.assertEqual(manager.all_steps(), [0, 1, 2] if steps_found else [])

      # create a new manager
      manager2 = LocalCheckpointManager(
          directory=self.local_directory,
          global_mesh=self.global_mesh,
          options=options,
      )
      self.assertEqual(manager2.all_steps(), [0, 1, 2] if steps_found else [])

    def test_local_look_for_all_partial_slice(self):
      p1 = self.global_mesh.devices[0][0].process_index
      p2 = self.global_mesh.devices[1][-1].process_index
      empty_process = [p1, p2]

      # steps for processes:
      # [2, -1], [1, 2]
      # [1, 2], [2, -1]
      # end result should be 2 since it's the only common step
      self.save_local_all_and_remove_partial(False, empty_process)

    def test_local_look_for_all_one_full_slice(self):
      empty_process = [0]

      # steps for processes:
      # [2, -1], [1, 2]
      # [1, 2], [1, 2]
      # end result should be [1, 2] since it's the common step in the bottom
      # slice.
      self.save_local_all_and_remove_partial(True, empty_process)

    def save_local_all_and_remove_partial(self, keep, empty_process):
      """Test case."""
      options = CheckpointManagerOptions(
          local=LocalCheckpointOptions(
              save_interval_steps=1,
              max_to_keep=2,
              debug_use_full_global_mesh=True,
          ),
      )
      manager = LocalCheckpointManager(
          directory=self.local_directory,
          global_mesh=self.global_mesh,
          options=options,
      )

      manager.save(
          0,
          args=args_lib.Composite(
              state=PyTreeSaveArgs(self.pytree),
              process_metadata=process_metadata_checkpoint_handler.ProcessMetadataSaveArgs(
                  global_mesh=self.global_mesh
              ),
          ),
      )
      manager.wait_until_finished()

      manager.save(
          1,
          args=args_lib.Composite(
              state=PyTreeSaveArgs(self.doubled_pytree),
              process_metadata=process_metadata_checkpoint_handler.ProcessMetadataSaveArgs(
                  global_mesh=self.global_mesh
              ),
          ),
      )
      manager.wait_until_finished()

      if multihost.process_index() in empty_process:
        test_utils.empty_directory(self.local_directory)

      manager.reload()

      manager.save(
          2,
          args=args_lib.Composite(
              state=PyTreeSaveArgs(self.pytree),
              process_metadata=process_metadata_checkpoint_handler.ProcessMetadataSaveArgs(
                  global_mesh=self.global_mesh
              ),
          ),
      )
      manager.wait_until_finished()

      self.assertEqual(manager.all_steps(), [1, 2] if keep else [2])

      # create a new manager
      manager2 = LocalCheckpointManager(
          directory=self.local_directory,
          global_mesh=self.global_mesh,
          options=options,
      )

      all_step = manager2.all_steps()

      self.assertEqual(all_step, [1, 2] if keep else [2])

    @parameterized.parameters(
        (1, []), (1, [0]), (None, [0, 3]), (None, [0, 1, 2, 3])
    )
    def test_local_latest_step(self, expected, empty_process):
      """Test case."""
      options = CheckpointManagerOptions(
          local=LocalCheckpointOptions(
              save_interval_steps=1,
              max_to_keep=1,
              debug_use_full_global_mesh=True,
          ),
          persistent=PersistentCheckpointOptions(
              save_interval_steps=1, max_to_keep=1
          ),
      )
      manager = LocalCheckpointManager(
          directory=self.local_directory,
          global_mesh=self.global_mesh,
          options=options,
      )

      manager.save(
          0,
          args=args_lib.Composite(
              state=PyTreeSaveArgs(self.pytree),
              process_metadata=process_metadata_checkpoint_handler.ProcessMetadataSaveArgs(
                  global_mesh=self.global_mesh
              ),
          ),
      )
      manager.save(
          1,
          args=args_lib.Composite(
              state=PyTreeSaveArgs(self.doubled_pytree),
              process_metadata=process_metadata_checkpoint_handler.ProcessMetadataSaveArgs(
                  global_mesh=self.global_mesh
              ),
          ),
      )
      manager.wait_until_finished()

      # confirm all hosts have latest step as 1
      latest_step = manager.latest_step()
      self.assertEqual(latest_step, 1)

      if multihost.process_index() in empty_process:
        test_utils.empty_directory(self.local_directory)
      manager.reload()

      self.assertEqual(manager.latest_step(), expected)

      # create a new manager
      manager2 = LocalCheckpointManager(
          directory=self.local_directory,
          global_mesh=self.global_mesh,
          options=options,
      )

      latest_step = manager2.latest_step()

      self.assertEqual(latest_step, expected)

    @parameterized.parameters((1, [3]), (2, [2, 3]))
    def test_local_latest_step_cached(self, max_to_keep, expected_steps):
      """Test case."""
      options = CheckpointManagerOptions(
          local=LocalCheckpointOptions(
              save_interval_steps=1,
              max_to_keep=max_to_keep,
              debug_use_full_global_mesh=True,
          )
      )
      manager = LocalCheckpointManager(
          directory=self.local_directory,
          global_mesh=self.global_mesh,
          options=options,
      )
      self.assertEmpty(manager._steps)

      manager.save(
          0,
          args=args_lib.Composite(
              state=PyTreeSaveArgs(self.pytree),
              process_metadata=process_metadata_checkpoint_handler.ProcessMetadataSaveArgs(
                  global_mesh=self.global_mesh
              ),
          ),
      )
      manager.wait_until_finished()

      # confirm all hosts have latest step as 0
      self.assertEqual(manager._steps, [0])
      self.assertEqual(manager.latest_step(), 0)

      test_utils.empty_directory(self.local_directory)

      self.assertEqual(manager.latest_step(), 0)
      manager.reload()
      # after reload, latest_step cached should be false
      self.assertEmpty(manager._steps)

      # latest_step should be None and cached after calling latest_step
      self.assertIsNone(manager.latest_step())
      self.assertEqual(manager._steps, [])

      manager.save(
          1,
          args=args_lib.Composite(
              state=PyTreeSaveArgs(self.doubled_pytree),
              process_metadata=process_metadata_checkpoint_handler.ProcessMetadataSaveArgs(
                  global_mesh=self.global_mesh
              ),
          ),
      )
      manager.save(
          2,
          args=args_lib.Composite(
              state=PyTreeSaveArgs(self.pytree),
              process_metadata=process_metadata_checkpoint_handler.ProcessMetadataSaveArgs(
                  global_mesh=self.global_mesh
              ),
          ),
      )
      manager.save(
          3,
          args=args_lib.Composite(
              state=PyTreeSaveArgs(self.doubled_pytree),
              process_metadata=process_metadata_checkpoint_handler.ProcessMetadataSaveArgs(
                  global_mesh=self.global_mesh
              ),
          ),
      )
      manager.wait_until_finished()

      if multihost.process_index() == 0:
        test_utils.empty_directory(self.local_directory)
      # still should be cache hit
      self.assertEqual(manager._steps, expected_steps)
      self.assertEqual(manager.latest_step(), 3)

      manager = LocalCheckpointManager(
          directory=self.local_directory,
          global_mesh=self.global_mesh,
          options=options,
      )
      self.assertEqual(manager._steps, expected_steps)
      self.assertEqual(manager.latest_step(), 3)


class CheckpointManagerTestBase:
  """Test case."""

  @barrier_compatible_test
  class Test(parameterized.TestCase):
    """Test case."""

    def make_global_mesh(
        self, replica_axis_index: int = 0
    ) -> jax.sharding.Mesh:
      raise NotImplementedError()

    def setUp(self):
      super().setUp()
      self.enter_context(
          flagsaver.flagsaver(
              experimental_orbax_use_distributed_process_id=True
          )
      )
      if not multihost.is_runtime_to_distributed_ids_initialized():
        multihost.initialize_runtime_to_distributed_ids()
        multihost.initialize_distributed_to_device_ids()

      # make sure each process is working on different directories
      self.local_directory = epath.Path(
          self.create_tempdir(
              name=f'local_checkpointing_test_pid{multihost.process_index()}'
          ).full_path
      )
      # We use the same path for the persistent directory across all processes
      # but create it only on the primary host. self.create_tempdir() uses
      # self._get_tempdir_path_test() thus path remains same across all
      # processes.
      self.persistent_directory = (
          epath.Path(self._get_tempdir_path_test())
          / 'persistent_checkpointing_test'
      )
      if multihost.is_primary_host(primary_host=0):
        self.persistent_directory = epath.Path(
            self.create_tempdir(name='persistent_checkpointing_test').full_path
        )
      logging.info(
          'self.local_directory=%s, self.persistent_directory=%s',
          self.local_directory,
          self.persistent_directory,
      )

      test_utils.set_tensorstore_driver_for_test()
      test_utils.sync_global_processes('CheckpointManagerTest:setup_complete')

    def tearDown(self):
      super().tearDown()
      test_utils.sync_global_processes(
          'CheckpointManagerTest:teardown_complete'
      )

    def setup_pytree(self, global_mesh: jax.sharding.Mesh):
      """Create mesh and pytree."""
      pytree = {
          'a': test_utils.create_sharded_array(
              np.arange(8), global_mesh, jax.sharding.PartitionSpec(None)
          ),
          'b': test_utils.create_sharded_array(
              np.arange(16),
              global_mesh,
              jax.sharding.PartitionSpec('data'),
          ),
          'scalar': test_utils.create_sharded_array(
              123, global_mesh, jax.sharding.PartitionSpec()
          ),
          'empty_dict': {},
          'empty_node': optax.EmptyState(),
      }
      return global_mesh, pytree

    @parameterized.parameters(
        (0, 1, 1, True),
        (0, 2, 2, True),
        (1, 1, 1, True),
        (1, 1, 2, True),
        (1, 2, 2, False),
        (3, 5, 7, False),
        (2, 1, 2, True),
        (3, 2, 3, True),
    )
    def test_should_save(
        self,
        step: int,
        local_interval: int,
        persistent_interval: int,
        expectation: bool,
    ):
      """Test case."""
      global_mesh, pytree = self.setup_pytree(self.make_global_mesh())
      abstract_state = jax.tree.map(utils.to_shape_dtype_struct, pytree)
      options = CheckpointManagerOptions(
          local=LocalCheckpointOptions(
              save_interval_steps=local_interval, max_to_keep=1
          ),
          persistent=PersistentCheckpointOptions(
              save_interval_steps=persistent_interval, max_to_keep=4
          ),
      )
      manager = CheckpointManager(
          local_directory=self.local_directory,
          persistent_directory=self.persistent_directory,
          global_mesh=global_mesh,
          abstract_state=abstract_state,
          options=options,
      )

      for i in range(step):
        manager.save(
            i,
            args=get_composite_save_args(pytree),
        )
        manager.wait_until_finished()

      self.assertEqual(manager.should_save(step), expectation)

    @parameterized.parameters(
        ([-1, -1, -1, -1], -1),
        ([0, 0, 0, -1], 0),
        ([2, 1, 0, -90000], 2),
        ([0, 1, 2, 3], 3),
        ([0, 0, 0, 0], 0),
        ([[0, 0], [1, 1], [2, 2], [3, 3]], [3, 3]),
        ([[0, -1], [1, 2], [0, 2], [-1, 2]], [1, 2]),
        ([False, False, False, False], 0),
        ([False, False, False, True], 1),
        ([False, True, True, True], 1),
        ([True, True, True, True], 1),
    )
    def test_global_max(self, inputs, expectation):
      """Test case."""
      local_host_inputs = inputs[multihost.process_index()]
      if not isinstance(local_host_inputs, list):
        local_host_inputs = [local_host_inputs]
        expectation = [expectation]
      self.assertEqual(
          checkpoint_manager._global_max(
              local_host_inputs, checkpoint_manager._get_global_broadcast_fn()
          ),
          expectation,
      )

    @parameterized.parameters(
        (1, 2, 2, 4, 10, [2, 4, 6, 8, 9]),
        (1, 3, 2, 4, 10, [2, 4, 6, 7, 8, 9]),
        (1, 2, 2, 5, 10, [0, 2, 4, 6, 8, 9]),
        (1, 2, 10, 4, 10, [0, 8, 9]),
    )
    def test_all_steps(
        self,
        local_interval,
        local_max_to_keep,
        persistent_interval,
        persistent_max_to_keep,
        total_steps,
        expectation,
    ):
      """Test case."""
      global_mesh, pytree = self.setup_pytree(self.make_global_mesh())
      abstract_state = jax.tree.map(utils.to_shape_dtype_struct, pytree)
      options = CheckpointManagerOptions(
          local=LocalCheckpointOptions(
              save_interval_steps=local_interval, max_to_keep=local_max_to_keep
          ),
          persistent=PersistentCheckpointOptions(
              save_interval_steps=persistent_interval,
              max_to_keep=persistent_max_to_keep,
          ),
      )
      manager = CheckpointManager(
          local_directory=self.local_directory,
          persistent_directory=self.persistent_directory,
          global_mesh=global_mesh,
          abstract_state=abstract_state,
          options=options,
      )

      for i in range(total_steps):
        manager.save(i, args=get_composite_save_args(pytree))
        manager.wait_until_finished()

      self.assertEqual(sorted(manager.all_steps()), expectation)

      # create a new manager
      manager2 = CheckpointManager(
          local_directory=self.local_directory,
          persistent_directory=self.persistent_directory,
          global_mesh=global_mesh,
          abstract_state=abstract_state,
          options=options,
      )

      self.assertEqual(sorted(manager2.all_steps()), expectation)

    @parameterized.parameters(
        (2, 2, 2, [0, 2, 4, 6, 8, 9]),
        (3, 2, 4, [0, 4, 7, 8, 9]),
        (2, 2, 5, [0, 8, 9]),
        (2, 6, 3, [0, 6, 8, 9]),
    )
    def test_all_steps_with_keep_interval(
        self,
        local_max_to_keep,
        persistent_interval,
        persistent_keep_period,
        expectation,
    ):
      """Test case."""
      global_mesh, pytree = self.setup_pytree(self.make_global_mesh())
      abstract_state = jax.tree.map(utils.to_shape_dtype_struct, pytree)
      options = CheckpointManagerOptions(
          local=LocalCheckpointOptions(
              save_interval_steps=1, max_to_keep=local_max_to_keep
          ),
          persistent=PersistentCheckpointOptions(
              save_interval_steps=persistent_interval,
              keep_period=persistent_keep_period,
              max_to_keep=1,
          ),
      )
      manager = CheckpointManager(
          local_directory=self.local_directory,
          persistent_directory=self.persistent_directory,
          global_mesh=global_mesh,
          abstract_state=abstract_state,
          options=options,
      )

      for i in range(10):
        manager.save(i, args=get_composite_save_args(pytree))
        manager.wait_until_finished()

      logging.info('all_steps: %s', manager.all_steps(True))
      manager.reload()
      logging.info('all_steps: %s', manager.all_steps(True))
      self.assertEqual(sorted(manager.all_steps(True)), expectation)
      manager.restore(8)
      manager.restore(9)

    @parameterized.parameters(
        (1, 4, 3),
        (2, 3, 3),
        (2, 4, 2),
        (4, 4, 0),
    )
    def test_latest_step(
        self,
        local_interval,
        persistent_interval,
        expectation,
        total_steps: Optional[int] = 4,
    ):
      """Test case."""
      global_mesh, pytree = self.setup_pytree(self.make_global_mesh())
      abstract_state = jax.tree.map(utils.to_shape_dtype_struct, pytree)
      options = CheckpointManagerOptions(
          local=LocalCheckpointOptions(
              save_interval_steps=local_interval, max_to_keep=2
          ),
          persistent=PersistentCheckpointOptions(
              save_interval_steps=persistent_interval, max_to_keep=4
          ),
      )
      manager = CheckpointManager(
          local_directory=self.local_directory,
          persistent_directory=self.persistent_directory,
          global_mesh=global_mesh,
          abstract_state=abstract_state,
          options=options,
      )

      for i in range(total_steps):
        manager.save(
            i,
            args=get_composite_save_args(pytree),
        )
        manager.wait_until_finished()

      self.assertEqual(manager.latest_step(), expectation)

      # create a new manager
      manager2 = CheckpointManager(
          local_directory=self.local_directory,
          persistent_directory=self.persistent_directory,
          global_mesh=global_mesh,
          abstract_state=abstract_state,
          options=options,
      )

      self.assertEqual(manager2.latest_step(), expectation)

    @parameterized.parameters(
        (False,),
        (True,),
    )
    def test_save(self, enable_async_checkpointing: bool):
      """Test case."""
      global_mesh, pytree = self.setup_pytree(self.make_global_mesh())
      abstract_state = jax.tree.map(utils.to_shape_dtype_struct, pytree)
      options = CheckpointManagerOptions(
          local=LocalCheckpointOptions(save_interval_steps=3, max_to_keep=2),
          persistent=PersistentCheckpointOptions(
              save_interval_steps=5, max_to_keep=3
          ),
          enable_async_checkpointing=enable_async_checkpointing,
      )
      manager = CheckpointManager(
          local_directory=self.local_directory,
          persistent_directory=self.persistent_directory,
          global_mesh=global_mesh,
          abstract_state=abstract_state,
          options=options,
      )

      for i in range(17):
        manager.save(
            i,
            args=get_composite_save_args(pytree),
        )
      manager.wait_until_finished()

      self.assertEqual(manager.all_steps(), [5, 10, 12, 15])

      with self.assertRaises(ValueError):
        manager.save(
            18,
            args=args_lib.Composite(
                state=PyTreeSaveArgs(pytree),
                process_metadata=process_metadata_checkpoint_handler.ProcessMetadataSaveArgs(
                    global_mesh=global_mesh
                ),
            ),
        )

    @parameterized.parameters(
        (False, 0),
        (True, 0),
        (True, 1),
    )
    def test_local_restore_by_broadcast(
        self, enable_async_checkpointing: bool, replica_axis_index: int
    ):
      """Test restore successfully."""
      global_mesh, pytree = self.setup_pytree(
          self.make_global_mesh(replica_axis_index)
      )
      abstract_state = jax.tree.map(utils.to_shape_dtype_struct, pytree)
      pytree_double = test_utils.apply_function(pytree, lambda x: x * 2)
      options = CheckpointManagerOptions(
          local=LocalCheckpointOptions(save_interval_steps=1, max_to_keep=2),
          persistent=PersistentCheckpointOptions(
              save_interval_steps=1, max_to_keep=2
          ),
          enable_async_checkpointing=enable_async_checkpointing,
          replica_axis_index=replica_axis_index,
      )

      with mock.patch.object(
          checkpoint_manager._MultisliceCheckpointManager,
          '_restore_from_persistent',
          autospec=True,
          return_value=None,
      ) as pcm_restore, CheckpointManager(
          local_directory=self.local_directory,
          persistent_directory=self.persistent_directory,
          global_mesh=global_mesh,
          options=options,
          abstract_state=abstract_state,
      ) as manager:

        manager.save(
            0,
            args=get_composite_save_args(pytree),
        )
        manager.save(
            1,
            args=get_composite_save_args(pytree),
        )
        manager.save(
            2,
            args=get_composite_save_args(pytree_double),
        )
        manager.wait_until_finished()

        if multihost.process_index() == 0:
          test_utils.empty_directory(self.persistent_directory)
        restored = manager.restore(2)
        test_utils.assert_tree_equal(self, pytree_double, restored)
        pcm_restore.assert_not_called()

    @parameterized.parameters(
        (0,),
        (1,),
    )
    def test_slice_setup(self, replica_axis_index: int):
      """Test restoration after slice swap (via global mesh)."""
      global_mesh, pytree = self.setup_pytree(
          self.make_global_mesh(replica_axis_index)
      )
      abstract_state = jax.tree.map(utils.to_shape_dtype_struct, pytree)
      pytree_double = test_utils.apply_function(pytree, lambda x: x * 2)
      options = CheckpointManagerOptions(
          local=LocalCheckpointOptions(save_interval_steps=1, max_to_keep=2),
          persistent=PersistentCheckpointOptions(
              save_interval_steps=1, max_to_keep=2
          ),
          replica_axis_index=replica_axis_index,
      )

      with CheckpointManager(
          local_directory=self.local_directory,
          persistent_directory=self.persistent_directory,
          global_mesh=global_mesh,
          options=options,
          abstract_state=abstract_state,
      ) as manager:
        self.assertEqual(
            manager._checkpoint_manager._persistent_primary_host, 0
        )
        self.assertEqual(manager._checkpoint_manager._local_primary_host, 2)
        slice_id = multislice.process_slice_id(
            multihost.process_index(),
            global_mesh,
            replica_axis_index=replica_axis_index,
        )
        self.assertEqual(slice_id == 0, manager.in_primary_slice)
        manager.save(
            0,
            args=get_composite_save_args(pytree),
        )
        manager.save(
            1,
            args=get_composite_save_args(pytree),
        )
        manager.save(
            2,
            args=get_composite_save_args(pytree_double),
        )

      self.assertEqual(jax.process_count(), 4)
      self.assertEqual(
          multislice.process_slice_id(
              0, global_mesh, replica_axis_index=replica_axis_index
          ),
          0,
      )
      self.assertEqual(
          multislice.process_slice_id(
              1, global_mesh, replica_axis_index=replica_axis_index
          ),
          0,
      )
      self.assertEqual(
          multislice.process_slice_id(
              2, global_mesh, replica_axis_index=replica_axis_index
          ),
          1,
      )
      self.assertEqual(
          multislice.process_slice_id(
              3, global_mesh, replica_axis_index=replica_axis_index
          ),
          1,
      )
      new_global_mesh = swap_slices_in_mesh(
          global_mesh, replica_axis_index=replica_axis_index
      )
      self.assertEqual(
          multislice.process_slice_id(
              0, new_global_mesh, replica_axis_index=replica_axis_index
          ),
          1,
      )
      self.assertEqual(
          multislice.process_slice_id(
              1, new_global_mesh, replica_axis_index=replica_axis_index
          ),
          1,
      )
      self.assertEqual(
          multislice.process_slice_id(
              2, new_global_mesh, replica_axis_index=replica_axis_index
          ),
          0,
      )
      self.assertEqual(
          multislice.process_slice_id(
              3, new_global_mesh, replica_axis_index=replica_axis_index
          ),
          0,
      )

    @parameterized.parameters(
        (0,),
        (1,),
    )
    def test_slice_swap_check_steps(self, replica_axis_index: int):
      """Test step detection across slice swap (via mesh)."""
      global_mesh, pytree = self.setup_pytree(
          self.make_global_mesh(replica_axis_index)
      )
      abstract_state = jax.tree.map(utils.to_shape_dtype_struct, pytree)
      options = CheckpointManagerOptions(
          local=LocalCheckpointOptions(save_interval_steps=1, max_to_keep=10),
          persistent=PersistentCheckpointOptions(
              save_interval_steps=10, max_to_keep=10
          ),
          replica_axis_index=replica_axis_index,
      )

      with CheckpointManager(
          local_directory=self.local_directory,
          persistent_directory=self.persistent_directory,
          global_mesh=global_mesh,
          options=options,
          abstract_state=abstract_state,
      ) as manager:
        self.assertEqual(
            manager._checkpoint_manager._persistent_primary_host, 0
        )
        self.assertEqual(manager._checkpoint_manager._local_primary_host, 2)
        slice_id = multislice.process_slice_id(
            multihost.process_index(),
            global_mesh,
            replica_axis_index=replica_axis_index,
        )
        self.assertEqual(slice_id == 0, manager.in_primary_slice)
        manager.save(
            0,
            args=get_composite_save_args(pytree),
        )
        manager.save(
            1,
            args=get_composite_save_args(pytree),
        )
        manager.save(
            2,
            args=get_composite_save_args(pytree),
        )
        manager.wait_until_finished()
        self.assertSameElements([0, 1, 2], manager.all_steps())
        self.assertEqual(2, manager.latest_step())

        # Delete the checkpoint on the *current* primary slice (it will
        # become the secondary below, and the only checkpoint available will be
        # on the primary slice).
        if manager.in_primary_slice:
          test_utils.empty_directory(self.local_directory)
          # Increment the counter on the primary slice (it was already
          # incremented on the secondary slice during the save), otherwise when
          # we use the counter again below, it will use count 0 *again*, which
          # will fail, since different barrier processes are used for the
          # barrier.
          next(checkpoint_manager.local_all_steps_broadcast_counter)
        test_utils.sync_global_processes('sync_after_local_dir_empty')

      new_global_mesh = swap_slices_in_mesh(
          global_mesh, replica_axis_index=replica_axis_index
      )
      abstract_state = jax.tree.map(
          functools.partial(
              _replace_abstract_array_sharding_with_mesh,
              mesh=new_global_mesh,
          ),
          abstract_state,
      )

      with CheckpointManager(
          local_directory=self.local_directory,
          persistent_directory=self.persistent_directory,
          global_mesh=new_global_mesh,
          options=options,
          abstract_state=abstract_state,
      ) as manager:
        self.assertEqual(
            manager._checkpoint_manager._persistent_primary_host, 2
        )
        self.assertEqual(manager._checkpoint_manager._local_primary_host, 0)
        slice_id = multislice.process_slice_id(
            multihost.process_index(),
            new_global_mesh,
            replica_axis_index=replica_axis_index,
        )
        # It is able to recover the steps even though only the primary slice
        # has local checkpoints. This has caused problems in the past because
        # the primary ordinarily is not responsible for dealing with local
        # checkpoints.
        if manager.in_primary_slice:
          self.assertNotEmpty(list(self.local_directory.iterdir()))
        else:
          self.assertEmpty(list(self.local_directory.iterdir()))
        self.assertEqual(slice_id == 0, manager.in_primary_slice)
        self.assertSameElements([0, 1, 2], manager.all_steps())
        self.assertEqual(2, manager.latest_step())

    @parameterized.parameters(
        (False, 0),
        (True, 0),
        (True, 1),
    )
    def test_local_missing_checkpoint(
        self, enable_async_checkpointing: bool, replica_axis_index: int
    ):
      """Test restore successfully."""
      global_mesh, pytree = self.setup_pytree(
          self.make_global_mesh(replica_axis_index)
      )
      abstract_state = jax.tree.map(utils.to_shape_dtype_struct, pytree)
      pytree_double = test_utils.apply_function(pytree, lambda x: x * 2)
      options = CheckpointManagerOptions(
          local=LocalCheckpointOptions(save_interval_steps=1, max_to_keep=2),
          persistent=PersistentCheckpointOptions(
              save_interval_steps=2, max_to_keep=2
          ),
          enable_async_checkpointing=enable_async_checkpointing,
          replica_axis_index=replica_axis_index,
      )

      with CheckpointManager(
          local_directory=self.local_directory,
          persistent_directory=self.persistent_directory,
          global_mesh=global_mesh,
          abstract_state=abstract_state,
          options=options,
      ) as manager:
        manager.save(
            0,
            args=get_composite_save_args(pytree),
        )  # both
        manager.save(
            1,
            args=get_composite_save_args(pytree),
        )  # local
        manager.save(
            2,
            args=get_composite_save_args(pytree_double),
        )  # both
        manager.wait_until_finished()

        if not manager.in_primary_slice:
          test_utils.empty_directory(self.local_directory)
        restored = manager.restore(2)
        test_utils.assert_tree_equal(self, pytree_double, restored)

        # Test subsequent saves.

        # reload manager gives each process the correct view of existing local
        # checkpoints.
        manager.reload()

        manager.save(
            3,
            args=get_composite_save_args(pytree),
        )  # local
        manager.save(
            4,
            args=get_composite_save_args(pytree),
        )  # both
        manager.save(
            5,
            args=get_composite_save_args(pytree_double),
        )  # local
        manager.wait_until_finished()

        restored = manager.restore(5)
        test_utils.assert_tree_equal(self, pytree_double, restored)

    @parameterized.parameters(
        (False, 0),
        (True, 0),
        (True, 1),
    )
    def test_persistent_checkpoint_restore(
        self, enable_async_checkpointing: bool, replica_axis_index: int
    ):
      """Test case."""
      global_mesh, pytree = self.setup_pytree(
          self.make_global_mesh(replica_axis_index)
      )
      abstract_state = jax.tree.map(utils.to_shape_dtype_struct, pytree)
      pytree_double = test_utils.apply_function(pytree, lambda x: x * 2)
      options = CheckpointManagerOptions(
          local=LocalCheckpointOptions(save_interval_steps=1, max_to_keep=2),
          persistent=PersistentCheckpointOptions(
              save_interval_steps=2, max_to_keep=2
          ),
          enable_async_checkpointing=enable_async_checkpointing,
          replica_axis_index=replica_axis_index,
      )

      with CheckpointManager(
          local_directory=self.local_directory,
          persistent_directory=self.persistent_directory,
          global_mesh=global_mesh,
          abstract_state=abstract_state,
          options=options,
      ) as manager:
        manager.save(
            0,
            args=get_composite_save_args(pytree),
        )
        manager.save(
            1,
            args=get_composite_save_args(pytree),
        )
        manager.save(
            2,
            args=get_composite_save_args(pytree_double),
        )
        manager.wait_until_finished()

        # remove all the local directories
        test_utils.empty_directory(self.local_directory)

        restored = manager.restore(2, args=args_lib.Composite(
            **{_STATE_ITEM_NAME: PyTreeRestoreArgs()}
            ))
        test_utils.assert_tree_equal(self, pytree_double, restored)

    def test_process_index_metadata(self):
      """Test case."""
      global_mesh, pytree = self.setup_pytree(self.make_global_mesh())
      abstract_state = jax.tree.map(utils.to_shape_dtype_struct, pytree)
      with CheckpointManager(
          local_directory=self.local_directory,
          persistent_directory=self.persistent_directory,
          global_mesh=global_mesh,
          abstract_state=abstract_state,
      ) as mngr:
        metadata_folder = mesh_consistency.process_metadata_folder(
            self.local_directory / '0' / 'process_metadata'
        )
        self.assertFalse(metadata_folder.exists())

        mngr.save(
            0,
            args=get_composite_save_args(pytree),
        )
        mngr.wait_until_finished()

        if mngr.in_primary_slice:
          self.assertFalse(metadata_folder.exists())
        else:
          metadata_path = (
              metadata_folder
              / mesh_consistency._GLOBAL_PROCESS_METADATA_FILE_NAME
          )
          self.assertTrue(metadata_path.exists())
          contents = json.loads(metadata_path.read_text())
          self.assertListEqual(
              multihost.distributed_to_device_ids(),
              contents,
          )

          metadata_path = (
              metadata_folder / mesh_consistency._MESH_METADATA_FILE_NAME
          )
          self.assertTrue(metadata_path.exists())
          device_ids = json.loads(metadata_path.read_text())
          self.assertListEqual(
              device_ids,
              [int(id) for id in global_mesh.device_ids.flatten()],
          )

    def test_should_save_fn(self):
      """Test case."""
      local_expectation = [1, 5, 7, 8]
      persistent_expectation = [1, 2]

      def should_save_local(step: int, _) -> bool:
        return step in local_expectation

      def should_save_persistent(step: int, _) -> bool:
        return step in persistent_expectation

      max_to_keep = 10
      global_mesh, pytree = self.setup_pytree(self.make_global_mesh())
      abstract_state = jax.tree.map(utils.to_shape_dtype_struct, pytree)
      options = CheckpointManagerOptions(
          local=LocalCheckpointOptions(
              should_save_fn=should_save_local, max_to_keep=max_to_keep
          ),
          persistent=PersistentCheckpointOptions(
              should_save_fn=should_save_persistent, max_to_keep=max_to_keep
          ),
      )
      manager = CheckpointManager(
          local_directory=self.local_directory,
          persistent_directory=self.persistent_directory,
          global_mesh=global_mesh,
          abstract_state=abstract_state,
          options=options,
      )

      for i in range(max_to_keep):
        manager.save(
            i,
            args=get_composite_save_args(pytree),
        )
        manager.wait_until_finished()

      self.assertSameElements(
          manager.all_steps(), set(local_expectation + persistent_expectation)
      )

    @parameterized.parameters(
        ((1, 2), 0, np.asarray([[0, 1]])),
        ((2, 2), 0, np.asarray([[2, 3], [0, 1]])),
        ((3, 2), 0, np.asarray([[4, 5], [2, 3], [0, 1]])),
        ((4, 2), 0, np.asarray([[6, 7], [4, 5], [2, 3], [0, 1]])),
        ((2, 1), 1, np.asarray([[0], [1]])),
        ((2, 2), 1, np.asarray([[1, 0], [3, 2]])),
        ((2, 3), 1, np.asarray([[2, 1, 0], [5, 4, 3]])),
        ((2, 4), 1, np.asarray([[3, 2, 1, 0], [7, 6, 5, 4]])),
    )
    def test_swap_slices_in_mesh(
        self, mesh_shape, replica_axis_index, expected_result
    ):
      """Test slice swapping utility."""

      class FakeMesh:

        def __init__(self, devices, axis_names):
          self.devices = devices
          self.axis_names = axis_names
          self.shape = mesh_shape
          self.shape_tuple = tuple(mesh_shape)

      devices = np.arange(np.prod(mesh_shape)).reshape(mesh_shape)
      with mock.patch.object(jax.sharding, 'Mesh', wraps=FakeMesh):
        swapped_mesh = swap_slices_in_mesh(
            jax.sharding.Mesh(devices, None),  # pytype: disable=wrong-arg-types
            replica_axis_index=replica_axis_index,
        )
        swapped_devices = swapped_mesh.devices
        np.testing.assert_array_equal(swapped_devices, expected_result)



# Copyright 2025 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Step level checkpoint metadata.

It is different from `path.step.Metadata` and `metadata.StepMetadata`.

TODO(niketkb): Figure out how to merge this with `path.step.Metadata`,
`metadata.StepMetadata` and `save_decision_policy_lib.StepInfo`.
"""

from __future__ import annotations

import dataclasses
import datetime
from typing import Any, Callable, Sequence

from absl import logging
import jax
from orbax.checkpoint._src import threading as threading_lib

PyTree = Any


@dataclasses.dataclass
class CheckpointInfo:
  """Basic information relevant to a checkpoint step.

  Attributes:
    step: Step number.
    time: Time at which the checkpoint was saved. This should be treated as an
      approximate start time of the save operation.
    metrics: Metrics associated with the step, if provided.
  """

  step: int
  time: datetime.datetime
  metrics: PyTree | None

  def __post_init__(self):
    # Users may provide step as a jax.Array.
    if isinstance(self.step, jax.Array):
      self.step = int(self.step)

  def __str__(self) -> str:
    return f'Checkpoint[step={self.step} | time={self.time}]'

  def __eq__(self, other: CheckpointInfo) -> bool:
    return self.step == other.step and self.time == other.time

  def __hash__(self) -> int:
    return hash((self.step, self.time))


class CheckpointInfos:
  """Thread-safe list of `CheckpointInfo`.

  It does not gurantee thread-safety for individual CheckpointInfo.

  See this doc for container thread-safety:
  https://docs.python.org/3/faq/library.html#what-kinds-of-global-value-mutation-are-thread-safe
  """

  def __init__(
      self,
      checkpoint_infos: Sequence[CheckpointInfo] | None = None,
      timeout_sec=5.0,
  ):
    """Initializes thread-safe list of `CheckpointInfo`."""
    self._lock = threading_lib.TimeoutRLock(timeout_sec)
    with self._lock:
      self._checkpoint_infos: list[CheckpointInfo] = list(
          checkpoint_infos or []
      )

  def set(self, checkpoint_infos: Sequence[CheckpointInfo]) -> CheckpointInfos:
    """Sets `checkpoint_infos` to `self` and returns it."""
    with self._lock:
      self._checkpoint_infos = list(checkpoint_infos)
      logging.vlog(
          1,
          'CheckpointInfos.set: steps=%s',
          [info.step for info in self._checkpoint_infos],
      )
    return self

  # Support indexing.
  def __getitem__(self, index):
    return self._checkpoint_infos[index]

  def __iter__(self):
    with self._lock:
      for info in self._checkpoint_infos:
        yield info

  def __len__(self):
    return len(self._checkpoint_infos)

  def latest(self) -> CheckpointInfo | None:
    """Returns the latest CheckpointInfo if any."""
    with self._lock:
      return self._checkpoint_infos[-1] if self._checkpoint_infos else None

  def empty(self) -> bool:
    """Returns true if the list of checkpoint infos is empty."""
    return not self._checkpoint_infos

  def size(self) -> int:
    """Returns the number of CheckpointInfo."""
    return len(self._checkpoint_infos)

  def delete_if(self, select_fn: Callable[[CheckpointInfo], bool]) -> None:
    """Deletes all CheckpointInfo that match `select_fn`."""
    with self._lock:
      retained_infos = []
      deleted_steps = []
      for info in self._checkpoint_infos:
        if select_fn(info):
          deleted_steps.append(info.step)
        else:
          retained_infos.append(info)
      self.set(retained_infos)
      logging.vlog(
          1, 'CheckpointInfos.delete_if: deleted steps=%s', deleted_steps
      )

  def append(self, checkpoint_info: CheckpointInfo) -> None:
    """Appends a CheckpointInfo."""
    self._checkpoint_infos.append(checkpoint_info)

  def __repr__(self):
    with self._lock:
      steps = [info.step for info in self._checkpoint_infos]
      return f'CheckpointInfos(size={self.size()}, steps={steps})'



# Copyright 2025 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""AbstractCheckpointer."""

import abc
from typing import Any, Optional
from absl import logging
from etils import epath
from orbax.checkpoint import version


class AbstractCheckpointer(abc.ABC):
  """An interface allowing atomic save and restore for a single object.

  Typically, an implementation of this class should rely on a
  :py:class:`.CheckpointHandler` object, which type-specific logic can be
  delegated to. In this way, the Checkpointer can be used for many different
  types, while itself only handling common logic related to atomicity,
  synchronization, or asynchronous thread management.
  """

  def __new__(cls, *args, **kwargs):  # pylint: disable=unused-argument
    logging.log_first_n(logging.INFO, version.get_details(), n=1)
    return super().__new__(cls)

  @abc.abstractmethod
  def save(self, directory: epath.PathLike, *args, **kwargs):
    """Saves the given item to the provided directory.

    Args:
      directory: a path to which to save.
      *args: additional args to provide to the CheckpointHandler's save method.
      **kwargs: additional keyword args to provide to the CheckpointHandler's
        save method.
    """
    pass

  @abc.abstractmethod
  def restore(self, directory: epath.PathLike, *args, **kwargs) -> Any:
    """Restores from the provided directory.

    Delegates to underlying handler.

    Args:
      directory: a path to restore from.
      *args: additional args to provide to the CheckpointHandler's restore
        method.
      **kwargs: additional keyword args to provide to the CheckpointHandler's
        restore method.

    Returns:
      a restored object
    """
    pass

  def structure(self, directory: epath.PathLike) -> Optional[Any]:
    """DEPRECATED.

    The structure of the saved object at `directory`.

    Delegates to underlying handler.

    Args:
      directory: a path to a saved checkpoint.

    Returns:
      the object structure or None, if the underlying handler does not implement
      `structure`.
    """
    pass

  def metadata(self, directory: epath.PathLike) -> Optional[Any]:
    """Returns metadata about the saved item.

    Ideally, this is a cheap way to collect information about the checkpoint
    without requiring a full restoration.

    Args:
      directory: the directory where the checkpoint is located.

    Returns:
      item metadata
    """
    pass

  def close(self):
    """Closes the Checkpointer."""
    pass



# Copyright 2025 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Defines exported symbols for the namespace package `orbax.checkpoint`."""

# pylint: disable=g-importing-member, g-bad-import-order

import contextlib
import functools

from orbax.checkpoint.experimental import v1
from orbax.checkpoint import arrays
from orbax.checkpoint import aggregate_handlers
from orbax.checkpoint import args
from orbax.checkpoint import checkpoint_utils
from orbax.checkpoint import checkpointers
from orbax.checkpoint import checkpoint_managers
from orbax.checkpoint import handlers
from orbax.checkpoint import logging
from orbax.checkpoint import metadata
from orbax.checkpoint import msgpack_utils
from orbax.checkpoint import options
from orbax.checkpoint import path
from orbax.checkpoint import serialization
from orbax.checkpoint import transform_utils
from orbax.checkpoint import tree
from orbax.checkpoint import type_handlers
from orbax.checkpoint import utils
from orbax.checkpoint import version
# TODO(cpgaffney): Import the public multihost API.
from orbax.checkpoint._src.multihost import multihost
from orbax.checkpoint._src.path import step

from orbax.checkpoint.future import Future

from orbax.checkpoint.transform_utils import apply_transformations
from orbax.checkpoint.transform_utils import merge_trees
from orbax.checkpoint.transform_utils import RestoreTransform
from orbax.checkpoint.transform_utils import Transform

from orbax.checkpoint.abstract_checkpoint_manager import AbstractCheckpointManager
from orbax.checkpoint.checkpointers import *
from orbax.checkpoint.checkpoint_manager import CheckpointManager
from orbax.checkpoint.checkpoint_manager import AsyncOptions
from orbax.checkpoint.checkpoint_manager import CheckpointManagerOptions

from orbax.checkpoint._src.handlers.pytree_checkpoint_handler import RestoreArgs
from orbax.checkpoint._src.handlers.pytree_checkpoint_handler import ArrayRestoreArgs
from orbax.checkpoint._src.handlers.pytree_checkpoint_handler import SaveArgs

# Important handlers.
from orbax.checkpoint._src.handlers.checkpoint_handler import CheckpointHandler
from orbax.checkpoint._src.handlers.async_checkpoint_handler import AsyncCheckpointHandler

from orbax.checkpoint._src.handlers.array_checkpoint_handler import ArrayCheckpointHandler
from orbax.checkpoint._src.handlers.composite_checkpoint_handler import CompositeCheckpointHandler
from orbax.checkpoint._src.handlers.composite_checkpoint_handler import CompositeOptions
from orbax.checkpoint._src.handlers.json_checkpoint_handler import JsonCheckpointHandler
from orbax.checkpoint._src.handlers.proto_checkpoint_handler import ProtoCheckpointHandler
from orbax.checkpoint._src.handlers.pytree_checkpoint_handler import PyTreeCheckpointHandler
from orbax.checkpoint._src.handlers.random_key_checkpoint_handler import JaxRandomKeyCheckpointHandler
from orbax.checkpoint._src.handlers.random_key_checkpoint_handler import NumpyRandomKeyCheckpointHandler
from orbax.checkpoint._src.handlers.standard_checkpoint_handler import StandardCheckpointHandler

from orbax.checkpoint._src.handlers.handler_registration import CheckpointHandlerRegistry
from orbax.checkpoint._src.handlers.handler_registration import DefaultCheckpointHandlerRegistry

# This class should be regarded as internal-only, and may be removed without
# warning.
from orbax.checkpoint._src.handlers.base_pytree_checkpoint_handler import BasePyTreeCheckpointHandler

# Test utils.
from orbax.checkpoint import test_utils

from orbax.checkpoint._src.serialization.type_handlers import PLACEHOLDER

# A new PyPI release will be pushed everytime `__version__` is increased.
__version__ = version.__version__
del version



# Copyright 2025 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Compatibility wrapper to help leaf handlers to work as V0 type_handlers."""

import copy
from typing import Generic, Sequence, Tuple, cast, get_args
from absl import logging
from etils import epath
import jax
from jax import tree_util as jtu
import jax.numpy as jnp
import numpy as np
from orbax.checkpoint._src.futures import future
from orbax.checkpoint._src.metadata import sharding as sharding_metadata
from orbax.checkpoint._src.metadata import value as value_metadata
from orbax.checkpoint._src.serialization import type_handlers as type_handlers_v0
from orbax.checkpoint._src.serialization import types as types_v0
from orbax.checkpoint.experimental.v1._src.context import context as context_lib
from orbax.checkpoint.experimental.v1._src.path import types as path_types
from orbax.checkpoint.experimental.v1._src.serialization import array_leaf_handler
from orbax.checkpoint.experimental.v1._src.serialization import numpy_leaf_handler
from orbax.checkpoint.experimental.v1._src.serialization import scalar_leaf_handler
from orbax.checkpoint.experimental.v1._src.serialization import types
from orbax.checkpoint.experimental.v1._src.synchronization import synchronization
from orbax.checkpoint.experimental.v1._src.tree import types as tree_types


class _PathAwaitingCreation(path_types.PathAwaitingCreation):
  """Implementation of `PathAwaitingCreation` that awaits contracted signals."""

  def __init__(self, path: path_types.Path, operation_id: str):
    self._path = path
    self._operation_id = operation_id

  def __truediv__(
      self, other: path_types.PathAwaitingCreation | path_types.PathLike
  ) -> path_types.PathAwaitingCreation:
    if isinstance(other, path_types.PathAwaitingCreation):
      other = other.path
    return _PathAwaitingCreation(self._path / other, self._operation_id)

  async def await_creation(self) -> path_types.Path:
    await synchronization.await_contracted_signals(self._operation_id)
    return self._path

  @property
  def path(self) -> path_types.Path:
    return self._path


def _keypath_from_param_name(param_name: str) -> tree_types.PyTreeKeyPath:
  """Converts a param name to a PyTreeKeyPath.

  This is based on reversing of the name construction from tree/utils.py's
  param_name_from_keypath.

  Args:
    param_name: A string representing the parameter name.

  Returns:
    A PyTreeKeyPath representing the parameter name.
  """
  return tuple([jtu.GetAttrKey(s) for s in param_name.split('.')])


def _construct_serialization_param(
    value: types.Leaf,
    info: types_v0.ParamInfo,
) -> types.SerializationParam[types.Leaf]:
  return types.SerializationParam(
      keypath=_keypath_from_param_name(info.name),
      value=value,
  )


def _construct_serialization_context(
    info: types_v0.ParamInfo,
) -> types.SerializationContext:
  return types.SerializationContext(
      # TODO(dnlng): should use actual wait
      parent_dir=_PathAwaitingCreation(
          info.parent_dir, context_lib.get_context().operation_id()
      ),
      ts_context=info.ts_context,
      byte_limiter=info.byte_limiter,
  )


def _construct_deserialization_param(
    info: types_v0.ParamInfo,
    restore_args: types_v0.RestoreArgs,
) -> types.DeserializationParam[
    array_leaf_handler.AbstractArray
    | numpy_leaf_handler.AbstractNumpy
    | scalar_leaf_handler.AbstractScalar
    | None
]:
  """Constructs a DeserializationParam from a ParamInfo and RestoreArg."""

  logging.vlog(1, 'compatibility.py: restore_args: %s', restore_args)

  if restore_args.restore_type == np.ndarray:
    # Numpy type
    value = numpy_leaf_handler.NumpyShapeDtype(
        dtype=restore_args.dtype,
        shape=None,
    )
  elif isinstance(restore_args.restore_type, type) and issubclass(
      restore_args.restore_type, get_args(scalar_leaf_handler.Scalar)
  ):
    # Scalar type
    logging.vlog(1, 'Scalar restore_type set to: %s', restore_args.restore_type)
    value = restore_args.restore_type
  elif isinstance(restore_args, type_handlers_v0.ArrayRestoreArgs):
    # JAX Array type, construct value as jax.ShapeDtypeStruct.
    arg = cast(type_handlers_v0.ArrayRestoreArgs, restore_args)

    logging.info(
        'name: %s, ArrayRestoreArgs: %s, write_shape: %s',
        info.name,
        arg,
        info.write_shape,
    )

    if arg.mesh is not None and arg.mesh_axes is not None:
      sharding = jax.sharding.NamedSharding(arg.mesh, arg.mesh_axes)
    elif arg.sharding is not None:
      if isinstance(arg.sharding, sharding_metadata.ShardingMetadata):
        sharding = arg.sharding.to_jax_sharding()
      else:
        sharding = arg.sharding
    else:
      sharding = None

    if sharding is None:
      # it's a numpy type
      value = numpy_leaf_handler.NumpyShapeDtype(
          dtype=arg.dtype,
          shape=arg.shape,
      )
    else:
      # it's a jax.Array type
      value = jax.ShapeDtypeStruct(arg.shape, arg.dtype, sharding=sharding)
  elif info.write_shape is not None:
    # TODO(dnlng): this is needed due to write_shape is passed into the
    # metadata() call, and then returned metadata will include this write_shape
    # in it. We should refractor how this write_shape is populated from the v0
    # so that the compatibility wrapper doesn't only work specific for
    # ArrayLeafHandler.
    value = array_leaf_handler.ArrayMetadata(
        shape=None,
        dtype=None,
        sharding_metadata=None,
        storage_metadata=value_metadata.StorageMetadata(
            chunk_shape=None,
            write_shape=info.write_shape,
        ),
    )
    logging.vlog(1, 'ArrayMetadata as param.value: %s', value)
  elif (
      restore_args.restore_type in (None, int, float)
      or isinstance(restore_args.restore_type, (np.dtype, jnp.dtype))
      or issubclass(restore_args.restore_type, np.number)
  ):
    # scalar type
    value = restore_args.restore_type
  else:
    raise ValueError(
        'Unsupported restore_args: %s. Cannot construct DeserializationParam.'
        % restore_args
    )

  logging.vlog(1, 'deserialization_param.value: %r', value)

  return types.DeserializationParam(
      keypath=_keypath_from_param_name(info.name),
      value=value,
  )


def _construct_deserialization_context(
    info: types_v0.ParamInfo,
) -> types.DeserializationContext:
  return types.DeserializationContext(
      parent_dir=info.parent_dir,
      ocdbt_checkpoint=info.is_ocdbt_checkpoint,
      zarr3_checkpoint=info.use_zarr3,
      ts_context=info.ts_context,
      byte_limiter=info.byte_limiter,
  )


def _validate_serialization_infos(
    infos: Sequence[types_v0.ParamInfo],
) -> None:
  """Validates that all infos share the same properties."""
  info0 = infos[0]
  for info in infos[1:]:
    if (
        (info0.parent_dir != info.parent_dir)
        or (info0.ts_context != info.ts_context)
        or (info0.byte_limiter != info.byte_limiter)
    ):
      raise ValueError(
          'All infos must have the same parent_dir, ts_context, and'
          ' byte_limiter.'
      )

  # TODO(dnlng): Add validation for ocdbt & zarr3.


def _validate_deserialization_infos(
    infos: Sequence[types_v0.ParamInfo],
) -> None:
  """Validates that all infos share the same properties."""
  info0 = infos[0]
  for info in infos:
    if (
        info0.parent_dir != info.parent_dir
        or info0.is_ocdbt_checkpoint != info.is_ocdbt_checkpoint
        or info0.use_zarr3 != info.use_zarr3
        or info0.ts_context != info.ts_context
        or info0.byte_limiter != info.byte_limiter
    ):
      raise ValueError(
          'All infos must have the same parent_dir, is_ocdbt_checkpoint,'
          ' use_zarr3, ts_context, and byte_limiter.'
      )


def _convert_v1_metadata_to_v0(
    name: str,
    directory: epath.Path | None,
    metadata: array_leaf_handler.AbstractArray,
) -> value_metadata.Metadata:
  """Converts a v1 metadata to a v0 metadata."""
  if isinstance(metadata, array_leaf_handler.ArrayMetadata):
    metadata = cast(array_leaf_handler.ArrayMetadata, metadata)
    ret = value_metadata.ArrayMetadata(
        name=name,
        directory=directory,
        sharding=metadata.sharding_metadata,
        shape=metadata.shape,
        dtype=metadata.dtype,
        storage=metadata.storage_metadata,
    )
    logging.vlog(1, 'ArrayMetadata: %s', ret)
    return ret
  elif isinstance(metadata, numpy_leaf_handler.NumpyMetadata):
    metadata = cast(numpy_leaf_handler.NumpyMetadata, metadata)
    ret = value_metadata.ArrayMetadata(
        name=name,
        directory=directory,
        sharding=None,
        shape=metadata.shape,
        dtype=metadata.dtype,
        storage=metadata.storage_metadata,
    )
    logging.vlog(1, 'NumpyMetadata: %s', ret)
    return ret
  elif isinstance(metadata, type):
    # scalar type such as int, float or np.number
    ret = value_metadata.ScalarMetadata(
        name=name,
        directory=directory,
        dtype=metadata,
    )
    logging.vlog(1, 'ScalarMetadata: %s', ret)
    return ret
  elif isinstance(metadata, (np.dtype, jnp.dtype)):
    # scalar type as np.dtype instances
    ret = value_metadata.ScalarMetadata(
        name=name,
        directory=directory,
        dtype=metadata.type,
    )
    logging.vlog(1, 'ScalarMetadata: %s', ret)
    return ret
  else:
    logging.warning(
        'Unsupported metadata type: %s. Returning value_metadata.Metadata'
        ' name=%s.',
        type(metadata),
        name,
    )
    return value_metadata.Metadata(
        name=name,
        directory=directory,
    )


class CompatibleTypeHandler(
    types_v0.TypeHandler, Generic[types.Leaf, types.AbstractLeaf]
):
  """Compatibility wrapper to help leaf handlers to work as V0 type_handlers."""

  def __init__(
      self,
      leaf_handler: types.LeafHandler[types.Leaf, types.AbstractLeaf],
      typestr: str,
  ):
    self._leaf_handler = leaf_handler
    self._typestr = typestr

  def typestr(self) -> str:
    return self._typestr

  async def serialize(
      self,
      values: Sequence[types.Leaf],
      infos: Sequence[types_v0.ParamInfo],
      args: Sequence[types_v0.SaveArgs] | None = None,
  ) -> Sequence[future.Future]:
    _validate_serialization_infos(infos)

    params = []
    info0 = infos[0]
    for info, value in zip(infos, values):
      logging.vlog(1, 'info: %s', info)
      params.append(_construct_serialization_param(value, info))
    serialization_context = _construct_serialization_context(info0)
    serialization_task = await self._leaf_handler.serialize(
        params, serialization_context
    )

    async def _background_serialize():
      await serialization_task

    operation_id = context_lib.get_context().operation_id()

    return [
        future.CommitFuture(
            coro=_background_serialize(),
            operation_id=operation_id,
        )
    ]

  async def deserialize(
      self,
      infos: Sequence[types_v0.ParamInfo],
      args: Sequence[types_v0.RestoreArgs] | None = None,
  ) -> Sequence[types.Leaf]:
    _validate_deserialization_infos(infos)

    params = []
    if args is None:
      args = [types_v0.RestoreArgs()] * len(infos)

    for info, restore_arg in zip(infos, args):
      if logging.vlog_is_on(1):
        logging.vlog(1, 'info: %s', info)
        logging.vlog(1, 'restore_arg: %s', restore_arg)

      # TODO(dnlng): need to allow passing in _construct_deserialization_param
      # for different leaf handlers.
      params.append(_construct_deserialization_param(info, restore_arg))

    info0 = infos[0]
    deserialization_context = _construct_deserialization_context(info0)
    task = await self._leaf_handler.deserialize(params, deserialization_context)
    return await task

  async def metadata(
      self, infos: Sequence[types_v0.ParamInfo]
  ) -> Sequence[value_metadata.Metadata]:
    """Constructs object metadata from a stored parameter location.

    Args:
      infos: sequence of ParamInfo

    Returns:
      Sequence of Metadata for each provided ParamInfo.
    """
    _validate_deserialization_infos(infos)
    args = [types_v0.RestoreArgs()] * len(infos)
    params = []

    for info, restore_arg in zip(infos, args):
      logging.vlog(1, 'info: %s', info)
      params.append(_construct_deserialization_param(info, restore_arg))

    info0 = infos[0]
    deserialization_context = _construct_deserialization_context(info0)

    metadatas = await self._leaf_handler.metadata(
        params, deserialization_context
    )

    ret = []
    for info, metadata in zip(infos, metadatas):
      ret.append(_convert_v1_metadata_to_v0(info.name, info.path, metadata))
    return ret

  def memory_size(
      self, values: Sequence[types.Leaf]
  ) -> Sequence[Tuple[int, int]]:
    # this only works for leaf handler that based on V0 TypeHandlers and stored
    # it in self._leaf_handler._handler_impl.
    if hasattr(self._leaf_handler, '_handler_impl'):
      v0_handler = self._leaf_handler._handler_impl  # pylint: disable=protected-access

      return v0_handler.memory_size(values)

    raise NotImplementedError(
        'Cannot resolve memory_size for this v1 leaf handler, '
        f' {self._leaf_handler!r}.'
    )

  @property
  def _array_metadata_store(self):
    # as the array_metadata_store.resolve_array_metadata_store read the metadata
    # store directly from _array_metadata_store, this is to provide the same
    # interface.  Currently, this only works for LeafHandler that based on
    # V0 ArrayHandler.
    if hasattr(self._leaf_handler, '_handler_impl') and hasattr(
        self._leaf_handler._handler_impl, '_array_metadata_store'  # pylint: disable=protected-access
    ):
      return self._leaf_handler._handler_impl._array_metadata_store  # pylint: disable=protected-access
    else:
      logging.warning(
          'Cannot resolve _array_metadata_store for this v1 leaf handler: %r',
          self._leaf_handler,
      )
      return None


def get_compatible_type_handler_registry(
    context: context_lib.Context | None = None,
    type_handler_registry: type_handlers_v0.TypeHandlerRegistry | None = None,
) -> type_handlers_v0.TypeHandlerRegistry:
  """Returns a V0 type handler registry that using v1 leaf handlers.

  This is a helper function to setup a v0 type handler registry that will be
  registered all existing v1 leaf handlers.

  Args:
    context: The context to that will be passed to create leaf handlers.
    type_handler_registry: The v0 type handler registry to use. If not provided,
      a new registry will be created.

  Returns:
    The v0 type handler registry that using standard v1 leaf handlers.
  """
  if type_handler_registry is None:
    type_handler_registry = copy.deepcopy(
        type_handlers_v0.GLOBAL_TYPE_HANDLER_REGISTRY
    )

  type_handler_registry.add(
      jax.Array,
      CompatibleTypeHandler(
          array_leaf_handler.ArrayLeafHandler(context=context),
          typestr=type_handlers_v0.JAX_ARRAY_TYPE_STR,
      ),
      override=True,
      ignore_warnings=True,
  )
  type_handler_registry.add(
      np.ndarray,
      CompatibleTypeHandler(
          numpy_leaf_handler.NumpyLeafHandler(context=context),
          typestr='np.ndarray',
      ),
      override=True,
      ignore_warnings=True,
  )

  for scalar_type in (int, float, bytes, np.number):
    type_handler_registry.add(
        scalar_type,
        CompatibleTypeHandler(
            scalar_leaf_handler.ScalarLeafHandler(context=context),
            typestr='scalar',
        ),
        override=True,
        ignore_warnings=True,
    )
  return type_handler_registry



# Copyright 2025 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utilities for parsing and validating checkpoint paths.

TODO(b/396190818): The validation functions in this module deserve to be
refactored to allow for greater flexibility, less specifically tied to one
implementation. At the moment they are tailored specifically to `save_pytree` /
`save_checkpointables`.
"""

import itertools

from absl import logging
from etils import epath
from orbax.checkpoint._src.metadata import checkpoint as checkpoint_metadata
from orbax.checkpoint._src.path import format_utils
from orbax.checkpoint._src.path import step as v0_step_lib
from orbax.checkpoint.experimental.v1._src.path import types as path_types


PYTREE_CHECKPOINTABLE_KEY = 'pytree'

METRICS_CHECKPOINTABLE_KEY = 'metrics'

RESERVED_CHECKPOINTABLE_KEYS = frozenset({
    METRICS_CHECKPOINTABLE_KEY,
})


def subdirs(directory: path_types.Path, *, limit: int = 3) -> list[str]:
  return list(
      itertools.islice(
          (subdir.name for subdir in directory.iterdir() if subdir.is_dir()),
          limit,
      )
  )


def validate_checkpoint_directory(path: path_types.PathLike):
  """Validates a checkpoint directory.

  Args:
    path: The path to the checkpoint directory.

  Raises:
    FileNotFoundError: If the path does not exist.
    NotADirectoryError: If the path is not a directory.
    ValueError: If the checkpoint is incomplete.
  """
  path = epath.Path(path)
  if not path.exists():
    raise FileNotFoundError(f'Checkpoint path {path} does not exist.')
  if not path.is_dir():
    raise NotADirectoryError(f'Checkpoint path {path} is not a directory.')
  if v0_step_lib.is_tmp_checkpoint(path):
    raise ValueError(f'Found incomplete checkpoint at {path}.')


def validate_checkpoint_metadata(path: path_types.PathLike):
  """Validates the checkpoint-level metadata (_CHECKPOINT_METADATA)."""
  metadata_store = checkpoint_metadata.metadata_store(enable_write=False)
  # Path points to a single step checkpoint with valid metadata.
  checkpoint_metadata_path = checkpoint_metadata.step_metadata_file_path(path)
  if not checkpoint_metadata_path.exists():
    raise FileNotFoundError(
        f'Checkpoint path {path} does not contain a valid metadata file:'
        f' {checkpoint_metadata_path.name}. Please ensure the path specified'
        ' for loading points to a valid Orbax checkpoint, indicated by the'
        ' presence of the metadata file.'
    )
  # TODO(niketkb): This check can be removed because the caller will anyway read
  # the metadata file. We are reading the file twice.
  if metadata_store.read(checkpoint_metadata_path) is None:
    raise ValueError(
        f'Failed to read valid metadata for checkpoint path {path}.'
    )


def validate_pytree_checkpoint(
    path: path_types.PathLike,
    *,
    checkpointable_name: str | None = PYTREE_CHECKPOINTABLE_KEY,
):
  """Validates a checkpoint path written by `ocp.save_pytree`.

  Args:
    path: The path to the checkpoint directory.
    checkpointable_name: The name of the checkpointable to load. A subdirectory
      with this name must exist in `directory`. If None then `directory` is
      expected to contain the checkpoint directly. Defaults to `pytree`.

  Raises:
    FileNotFoundError: If the path does not exist, or if `pytree` is not found
      in the directory
    ValueError: If the PyTree checkpoint is malformed.
  """
  path = epath.Path(path)
  pytree_dir = (
      path if checkpointable_name is None else path / checkpointable_name
  )
  if checkpointable_name is not None and not pytree_dir.exists():
    raise FileNotFoundError(
        f'Checkpoint path {path} must contain a subdirectory named'
        f' "{checkpointable_name}". Found subdirectories: {subdirs(path)}.'
        ' Please try inspecting the checkpointable metadata using'
        ' `ocp.checkpointables_metadata()` or try loading the checkpoint using'
        ' `ocp.load_checkpointables()`.'
    )
  if not format_utils._has_pytree_metadata_file(  # pylint: disable=protected-access
      pytree_dir
  ):
    # TODO(niketkb): Add following details to the error message:
    # 1. we should check other available subdirectories and see if any of them
    #   look like PyTree checkpoints, and instruct the user to consider
    #   whether they meant to specify any of those.
    # 2. we need to check the directory - if it contains PyTree files, suggest
    #   loading with checkpointable_name=None
    raise FileNotFoundError(
        f'Checkpoint path {path} does not contain a PyTree metadata file.'
    )
  if not format_utils._has_tensorstore_data_files(  # pylint: disable=protected-access
      pytree_dir
  ):
    logging.warning(
        'TensorStore data files not found in checkpoint path %s. This may be a'
        ' sign of a malformed checkpoint, unless your checkpoint consists'
        ' entirely of strings or other non-standard PyTree leaves.',
        path,
    )



# Copyright 2025 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Metadata describing checkpoints."""

from __future__ import annotations

import pprint
from typing import Any, Generic, TypeAlias, TypeVar

from orbax.checkpoint.experimental.v1._src.tree import types as tree_types


CheckpointableMetadataT = TypeVar('CheckpointableMetadataT')
SerializedMetadata = TypeVar('SerializedMetadata', bound=dict[str, Any])


PyTreeMetadata: TypeAlias = tree_types.PyTreeOf[tree_types.AbstractLeafType]
PyTreeMetadata.__doc__ = """
Metadata describing a PyTree checkpoint.

A serialized PyTree structure with the same structure as the checkpointed
PyTree. By "serialized", we mean that the PyTree has been converted to a
standardized representation, with all container nodes represented as standard
types (e.g., tuple, list, dict, etc.). The leaves of the tree are individual
parameter metadatas.
"""


class CheckpointMetadata(Generic[CheckpointableMetadataT]):
  """Represents complete metadata describing a checkpoint.

  Note that this class has a generic type `CheckpointableMetadataT`. This
  will typically be either :py:data:`.PyTreeMetadata` (see above), or
  `dict[str, Any]`.

  `CheckpointMetadata` can be accessed via one of two metadata methods. Please
  see `ocp.pytree_metadata` and `ocp.checkpointables_metadata` for more
  information and usage instructions.

  If the checkpoint contains a PyTree, this metadata can be acccessed via::

    metadata = ocp.pytree_metadata(path)

    # Inspect various properties
    metadata.init_timestamp_nsecs

    # Inspect the tree structure
    metadata.metadata.pytree
    metadata.metadata.pytree['layer0']['bias'].shape
    metadata.metadata.pytree['layer0']['bias'].dtype

  The checkpoint metadata can also be accessed more generically via::

    metadata = ocp.checkpointables_metadata(path)

    metadata.metadata.keys()  # == ['pytree', 'dataset', etc.]
    metadata.metadata['pytree']  # instance of PyTreeMetadata

  Attributes:
    metadata: Metadata for the checkpointable.
    init_timestamp_nsecs: timestamp when uncommitted checkpoint was initialized.
      Specified as nano seconds since epoch. default=None.
    commit_timestamp_nsecs: commit timestamp of a checkpoint, specified as nano
      seconds since epoch. default=None.
    custom_metadata: User-provided custom metadata. An arbitrary
      JSON-serializable dictionary the user can use to store additional
      information. The field is treated as opaque by Orbax.
  """

  def __init__(
      self,
      *,
      metadata: CheckpointableMetadataT,
      init_timestamp_nsecs: int | None = None,
      commit_timestamp_nsecs: int | None = None,
      custom_metadata: tree_types.JsonType | None = None,
  ):
    self._metadata = metadata
    self._init_timestamp_nsecs = init_timestamp_nsecs
    self._commit_timestamp = commit_timestamp_nsecs
    self._custom_metadata = custom_metadata

  @property
  def metadata(self) -> CheckpointableMetadataT:
    return self._metadata

  @property
  def init_timestamp_nsecs(self) -> int | None:
    return self._init_timestamp_nsecs

  @property
  def commit_timestamp_nsecs(self) -> int | None:
    return self._commit_timestamp

  @property
  def custom_metadata(self) -> tree_types.JsonType | None:
    return self._custom_metadata


  @classmethod
  def from_metadata(
      cls, metadata: CheckpointableMetadataT
  ) -> CheckpointMetadata[CheckpointableMetadataT]:
    return cls(metadata=metadata)

  def _properties_strings(self) -> dict[str, str]:
    return {
        'metadata': str(self.metadata),
        'init_timestamp_nsecs': str(self.init_timestamp_nsecs),
        'commit_timestamp_nsecs': str(self.commit_timestamp_nsecs),
        'custom_metadata': str(self.custom_metadata),
    }

  def __repr__(self):
    return f'CheckpointMetadata({pprint.pformat(self._properties_strings())})'



# Copyright 2025 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""StatefulCheckpointableHandler class."""

from typing import Any, Awaitable, Generic
from orbax.checkpoint.experimental.v1._src.handlers import types as handler_types
from orbax.checkpoint.experimental.v1._src.path import types as path_types

T = handler_types.T


class StatefulCheckpointableHandler(
    handler_types.CheckpointableHandler[T, T],
    Generic[T],
):
  """Serializes/deserializes a Checkpointable."""

  async def save(
      self,
      directory: path_types.PathAwaitingCreation,
      checkpointable: handler_types.StatefulCheckpointable[T],
  ) -> Awaitable[None]:
    return await checkpointable.save(directory)

  async def load(
      self,
      directory: path_types.Path,
      abstract_checkpointable: (
          handler_types.StatefulCheckpointable[T] | None
      ) = None,
  ) -> Awaitable[T]:
    if abstract_checkpointable is None:
      raise ValueError(
          'To restore a `StatefulCheckpointable`, you must pass an instance of'
          ' the object.'
      )

    # Returns Awaitable[None]
    background_load = await abstract_checkpointable.load(directory)

    async def _background_load() -> T:
      await background_load
      # After loading, `abstract_checkpointable` (actually just a concrete
      # checkpointable) should be populated with the loaded data.
      return abstract_checkpointable

    return _background_load()

  async def metadata(self, directory: path_types.Path) -> T:
    raise NotImplementedError(
        'Metadata retrieval is not supported for objects implementing'
        ' `StatefulCheckpointable`.'
    )

  def is_handleable(self, checkpointable: Any) -> bool:
    return isinstance(checkpointable, handler_types.StatefulCheckpointable)

  def is_abstract_handleable(self, abstract_checkpointable: Any) -> bool | None:
    return isinstance(
        abstract_checkpointable, handler_types.StatefulCheckpointable
    )



# Copyright 2025 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utilities for working with threads and asyncio event loops."""

import asyncio
import threading
from typing import Awaitable, Generic, TypeVar

T = TypeVar('T')


class BackgroundThreadRunner(Generic[T]):
  """A runner for an asyncio event loop in a background thread.

  This class expects an awaitable function that will be run in a background
  thread. It creates an event loop that is passed to the thread. This event loop
  should only be interacted with via asyncio thread-safe APIs, when tasks are
  scheduled from the main thread.
  """

  def __init__(
      self,
      target: Awaitable[T],
  ):
    self._target = target
    self._event_loop = asyncio.new_event_loop()
    self._thread = threading.Thread(
        target=self._event_loop_runner, args=(self._event_loop,), daemon=True
    )
    self._thread.start()
    self._future = asyncio.run_coroutine_threadsafe(
        self._target, self._event_loop
    )

  def _event_loop_runner(self, event_loop: asyncio.AbstractEventLoop):
    event_loop.run_forever()
    event_loop.close()

  def result(self, timeout: float | None = None) -> T:
    r = self._future.result(timeout=timeout)
    if self._thread:
      self._event_loop.call_soon_threadsafe(self._event_loop.stop)
      self._thread.join(timeout=timeout)
      self._thread = None
    return r
